// Copyright 2007 Tapio Auvinen, Erik Fallenius, Ville Karavirta,
// Jussi Nikander, and Harri Salonen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cexer;

import java.util.List;
import java.util.LinkedList;
import java.util.Date;
import java.util.Hashtable;
import java.io.Serializable;
import java.text.SimpleDateFormat;

public class Course implements Serializable {

    /**
     *  Id for serialization
     */
    private static final long serialVersionUID = 1L;
    
    // id must be unique
    private int id;
    private String code, name;
    private List rounds;
    private Hashtable names;
    private Hashtable descriptions;
    private Hashtable links;

    // minimum amount of points required
    private int min;

    // fraction of maximum points required
    private double min_frac;

    // Course-specific applet parameters
    private List additionalParameters;
    private Date open, close;

    /** Name of the class used to validate the student number. */
    private String checker;
    public static final String DEFAULT_CHECKER = "none";

    private static SimpleDateFormat df = new SimpleDateFormat("dd.MM.yyyy HH:mm");

    private int[] gradingMinPoints;
    private String[] gradingGrades;

    /** The responsible organizing person of the TRAKLA2-course, could be main assistant or lecturer. */
    private Student organizer;

    /** The organization that offers the course. */
    private Organization organization;


    public Course(String code) {
        this.code = code;
        names = new Hashtable();
        descriptions = new Hashtable();
        links = new Hashtable();
        rounds = new LinkedList();
        min = -1;
        min_frac = -1;
    }

    public Course(String code, String name) {
        this(code);
        this.name = name;
    }

    public String getCourseCode() {
        return code;
    }

    public void setCourseCode(String c) {
        code = c;
    }

    public String getCourseName() {
        return name;
    }

    public void setCourseName(String n) {
        name = n;
    }

    public String getFullCourseName() {
        return code + " " + name;
    }

    public void setOrganizer(Student organizer) {
        this.organizer = organizer;
    }

    public Student getOrganizer() {
        return organizer;
    }

    public void setOrganization(Organization organization) {
        this.organization = organization;
    }

    public Organization getOrganization() {
        return organization;
    }

    public void addRound(Round r) {
        rounds.add(r);
    }

    public void removeRound(Round r) {
        rounds.remove(r);
    }

    public List getRoundList() {
        return rounds;
    }


    // all in all a crappy implementation:
    // running time mn if there are m rounds which each have
    // approximately n exercises
    public int getMax() {
        int max = 0;
        List rounds = getRoundList();
        for(int i = 0; i < rounds.size(); i++) {
            max += ((Round)rounds.get(i)).getMax();
        }

        return max;
    }

    public int getId() {
        return id;
    }

    public void setId(int s) {
        id = s;
    }

    public void setAdditionalAppletParameters(List l) {
        additionalParameters = l;
    }

    public List getAdditionalAppletParameters() {
        return additionalParameters;
    }

    /**
     * Returns the minimum amount of points required to pass this course.
     * The amount of points required to pass a course is independent of the
     * amount of points to pass a single round.
     * The minimum points can be given either as an integer number representing
     * the minimum, or as a fraction of the maximum points. The minimum points
     * when a fraction is used is calculated by: max * frac
     *
     * An integer minimum must satisfy the following: 0 <= min <= max
     *
     * A fraction minimum must satisfy the following: 0 <= min <= 1
     *
     * An integer is preferred over fraction. If no valid value has been
     * given, 0 is returned.
     */
    public int getMin() {
        if (0 <= min && getMax() >= min) {
            return min;
        }

        if (0 <= min_frac && 1 >= min_frac) {
            return Math.round((float)(getMax() * min_frac));
        }

        return 0;
    }

    public void setMin(double value) {
        if(value > 0 && value < 1) {
            min_frac = value;
            return;
        }

        if(value >= 0) {
            min = Math.round((float)value);
            return;
        }

        if(value < 0) {
            min = -1;
            min_frac = -1;
        }
    }

    // Courses are equal if they have the same id
    public boolean equals(Object o) {
        if (!(o instanceof Course)) {
            return false;
        }

        return (getId()  == ((Course)o).getId());
    }

    public void addLocalizedData(String language, String name, String description, String link) {
        // XXX - just make the first added name as the default
        // to get getName() return something.  Better ideas??
        if (this.name == null) {
            this.name = name;
        }

        names.put(language, name);

        if (description == null) {
            description = "";
        }

        descriptions.put(language, description);

        if(link == null) {
            link="";
        }

        links.put(language, link);
    }

    public String getLocalizedName(String language) {
        if (names.containsKey(language)) {
            return (String)names.get(language);
        } else {
            return getCourseName();
        }
    }

    public String getLocalizedDescription(String language) {
        if (descriptions.containsKey(language)) {
            return (String)descriptions.get(language);
        } else {
            return "";
        }
    }

    public void setLocalizedDescription(String language, String description) {
        descriptions.put(language, description);
    }


    public String getLocalizedLink(String lang) {
        if(links.containsKey(lang))
            return (String)links.get(lang);
        return "";
    }

    public void setLocalizedLink(String lang, String link) {
        links.put(lang, link);
    }

    public void setOpen(Date o) {
        open = o;
    }

    public Date getOpen() {
        return open;
    }

    public String getFormattedOpen() {
        return df.format(open);
    }

    public void setClose(Date c) {
        close = c;
    }

    public Date getClose() {
        return close;
    }

    public String getFormattedClose() {
        return df.format(close);
    }

    /**
     * Returns the name of the class used to validate the student
     * number for this Course.
     * @see cexer.util.Checker
     */
    public String getCheckerClass() {
        if (checker == null) {
            String c = DataStorage.getInstance().getParameterValue("id_checker", "c", getId());
            if (c != null && !c.equals(""))
            checker = c;
            else
            checker = DEFAULT_CHECKER;
        }
        return checker;
    }

    /**
     * Sets the name of the class used to validate the student
     * number for this Course.
     * @see cexer.util.Checker
     */
    public void setCheckerClass(String value) {
        checker = value;
    }

    public void setGrades(int[] minPoints, String[] grade) {
        //XXX make sure the grades are sorted!! Now it is done in
        // DataStorage
        gradingMinPoints = minPoints;
        gradingGrades = grade;
    }

    public String[] getGrades() {
        return gradingGrades;
    }

    public int[] getGradeMinPoints() {
        return gradingMinPoints;
    }


    public String getGrade(int points) {
        if(gradingMinPoints == null || gradingGrades == null) {
            return "0";
        }

        int i = 0;
        String grade = "";

        while (i < gradingMinPoints.length && gradingMinPoints[i] <= points) {
            grade = gradingGrades[i];
            i++;
        }

        if(grade == "") {
            return "0";
        }

        return grade;
    }

    public String toString() {
        return code + " " + name;
    }

    public boolean isOpen() {
        Date now = new Date();
        if (now.after(getOpen()) && now.before(getClose()))
            return true;
        else
            return false;
    }

    public java.util.Iterator getSupportedLocaleIterator() {
        return names.keySet().iterator();
    }

    public List getSupportedLocales() {
        return new LinkedList(names.keySet());
    }

}
