// Copyright 2007 Tapio Auvinen, Erik Fallenius, Ville Karavirta,
// Jussi Nikander, and Harri Salonen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cexer;

import java.sql.*;
import cexer.util.*;
import cexer.xmlrpc.*;
import java.io.InputStream;
import java.io.IOException;
import java.net.BindException;
import java.util.*;
import java.text.*;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheException;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;

import org.apache.xmlrpc.WebServer;

/**
 * Database layer class for the TRAKLA2 application. Handles all communication
 * between the business layer and the database. Basic settings (like database name, user and password)
 * are read from the file <code>Datastorage.properties</code> at application startup.
 */
public class DataStorage {

    // Status codes used in Result table
    public static final int OK = 0;
    public static final int AFTER_DEADLINE = 1;
    public static final int INVALID_FINGERPRINT = 2;
    public static final int AFTER_CLOSING = 3; // Currently unused, for submissions after the course has closed
    public static final int ALL_SUBMISSIONS_USED = 4;
    
    // Status codes used in Result table when the points are modified from the monitor interface
    public static final int OK_MODIFIED = 10;
    public static final int AFTER_DEADLINE_MODIFIED = 11;
    public static final int INVALID_FINGERPRINT_MODIFIED = 12;
    public static final int AFTER_CLOSING_MODIFIED = 13; // Currently unused, for submissions after the course has closed
    public static final int ALL_SUBMISSIONS_USED_MODIFIED = 14;
    
        

    // Error codes when adding resutls
    public static final int STUDENT_NOT_FOUND = 21;
    public static final int COURSE_NOT_FOUND = 22;
    public static final int STUDENT_NOT_ON_COURSE = 23;
    public static final int DATABASE_ERROR = 24;
    public static final int ROUND_NOT_FOUND = 25;
    

    // User priviledges in the REGISTRATIONS table
    // priviledges for course staff MUST BE higher than
    // the value of STUDENT, and priviledges that mark either
    // the student's status on a course must be less or equal to STUDENT    

    public static final int MOVED = -2;
    public static final int UNREGISTERED = -1;
    public static final int STUDENT = 0;
    public static final int INSTRUCTOR = 1;
    public static final int ADMIN = 2;

    // Special Course IDs
    public static final int ALL_COURSES = 0;
    public static final int DUMMY_COURSE = Integer.MAX_VALUE;

    public static final double MILLIS_IN_HOUR = 3600000;
    public static final long LOCK_TIMEOUT = 1800000;
    //public static final long LOCK_TIMEOUT = 60000; // 10s timeout for testing purposes 

    // Exercise types
    public static final int UNKNOWN_EXERCISE = -1;
    public static final int APPLET_EXERCISE = 0;
    public static final int MULTIPLE_CHOICE_EXERCISE = 1;

    private HashMap locks;
    private Cache exerciseTemplateCache;
    private Cache courseCache;
    private Cache studentCache;
    //private CourseSelectionModel courseSelectionModel;
    //private TemplateSelectionModel templateSelectionModel;
    //private LocaleSelectionModel localeSelectionModel;

    private PreparedStatement preparedStatement;

    /**
     * Singleton DataStorage instance
     */
    private static DataStorage theInstance = null;


    /**
     * Gets the instance of the DataStorage class. If no instance exists,
     * a new is created.
     * 
     * @return DataStorage instance
     */
    public static synchronized DataStorage getInstance() {
        if(theInstance == null) {
            theInstance = new DataStorage();
        }

        return theInstance;
    }

    /**
     * Private constructor
     */
    private DataStorage() {
        locks = new HashMap();
        preparedStatement = null;
        initialize();
        Logging.info(this, "DataStorage initialized");
    }


    // ---- XXX Public interface XXX ----

    /**
     * Empty the system cache. This removes cached courses, students and exercises.
     */
    public void emptyCache() {
        try {
            courseCache.removeAll();
            studentCache.removeAll();
            exerciseTemplateCache.removeAll();
        } catch (CacheException e) {
            Logging.error(this, "Could not empty cache: " + e.getMessage());
            e.printStackTrace();
        } catch (IllegalStateException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }/*
        Shouldn't we upgrade to ehcache 1.3.0 as it seems to work? Older versions are a bit harder to find. -TA
        catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        */
    }

    /**
     * Get the results for a course
     * @param course
     * @return
     */
    public List getCourseResults(Course course) {
        Connection c = getDatabaseConnection();
        List l = getStudentResultsForCourseFromDatabase(course, c);

        closeDatabaseConnection(c);
        return l;
    }

    /**
     * Get the monitor object associated with the given username
     * @param userName
     * @return monitor object or null if the user name wasn't found in the db
     */
    public Student getMonitor(String userName) {
        Connection c = getDatabaseConnection();
        Student s = getMonitorFromDatabase(userName, c);
        if(s != null) {
            // a student instance that is a monitor must not be added to the cache,
            // since it will mess up the student's UI for the user. -jtn 8.9.05.
            //addStudentToCache(s);
            closeDatabaseConnection(c);
        }

        return s;
    }

    /** 
     * Returns the Student with the given id. If the student
     * does not exist in the database, returns null.
     * @param studentId Id of the student to fetch.
     * @return The student, or null if not found from db.
     */
    public Student getStudent(String studentId) {
        Element e = studentCache.get(studentId);
        Student s = null;

        if (e != null) {
            s = (Student) e.getValue();
        }

        if (s == null || !s.getId().equals(studentId)) {
            Connection c = getDatabaseConnection();
            s = getStudentFromDatabase(studentId, c);
            closeDatabaseConnection(c);
            if (s != null) {
                addStudentToCache(s);
            }
        }

        return s;
    }


    /**
     * Searches for a student on the courses defined with courses-parameter.
     * @param studentId Student ID
     * @param courses List of accepted courses
     * @return found student or otherwise null
     */
    public Student getStudent(String studentId, List courses) {
        Element e = studentCache.get(studentId);
        Student s = null;
        if (e != null) {
            s = (Student) e.getValue();
        }

        if(s == null || !s.getId().equals(studentId)) {
            Connection c = getDatabaseConnection();
            s = getStudentFromDatabase(studentId, c);
            closeDatabaseConnection(c);
            if (s != null) {
                addStudentToCache(s);
            }
        }

        Course acceptedCourse = null;
        if(s == null) {
            return null;
        }

        ListIterator acceptedCourseIterator = courses.listIterator();
        while(acceptedCourseIterator.hasNext()) {
            acceptedCourse = (Course)acceptedCourseIterator.next();
            if(s.isOnCourse(acceptedCourse)) {
                return s;
            }
        }

        return null;
    }


    /**
     * Get the course with the given course id
     * @param courseId
     * @return course
     */
    public synchronized Course getCourse(int courseId) {
        if (courseId < 1) {
            return null;
        }

        Element e = courseCache.get("" + courseId);
        Course c = null;
        if (e != null) {
            c = (Course)e.getValue();
        }

        if(c == null || c.getId() != courseId) {
            Connection con = getDatabaseConnection();
            if(courseId == DUMMY_COURSE) {
                c = new Course("","");
                c.setId(DUMMY_COURSE);
                List l = getAppletParametersFromDatabase(DUMMY_COURSE,'c',con);
                c.setAdditionalAppletParameters(l);
            } else {
                c = getCourseFromDatabase(courseId, con);
            }
            closeDatabaseConnection(con);
        }

        return c;
    }

    /**
     * Return a Statistics object for the course with the given id
     * @param courseId
     * @return
     */
    public Statistics getStatisticsForCourse(int courseId) {
        Connection con = getDatabaseConnection();
        Statistics s = getStatisticsFromDatabase(courseId, con);
        closeDatabaseConnection(con);
        return s;
    }

    /**
     * Adds the given student to the database. If the student does not have a
     * student number and password the student is not added. Also, if the student
     * exists, nothing is done.
     * @param s The student to add to the db.
     * @return True if inserted ok, false otherwise.
     */
    public boolean addStudent(Student s) {
        // student already in db
        if (getStudent(s.getId()) != null) {
            Logging.error(this, "addStudent(): student already in DB");
            return false;
        }

        // student number null or empty
        // We will let them in anyway and ask the studentnumber later. 
        /*
        if (s.getStudentNumber() == null || s.getStudentNumber().trim().equals("")) {
            Logging.error(this, "addStudent(): student number null");
            return false;
        }
        */

        // password null or empty
        //if (s.getPassword() == null || s.getPassword().trim().equals("")) {
        //    return false;
        //}

        try {
            Connection con = getDatabaseConnection();
            String ps = "INSERT INTO users(id, name, password, lang, email, studentNumber) "
                + "VALUES (?,?,?,?,?,?)";
            PreparedStatement st = con.prepareStatement(ps);
            st.setString(1, s.getId());
            st.setString(2, s.getName());
            st.setString(3, s.getPassword());
            st.setString(4, s.getLanguage());
            st.setString(5, s.getEmail());
            st.setString(6, s.getStudentNumber());
            st.executeUpdate();

            closeDatabaseConnection(con);
            addStudentToCache(s);

            Logging.info(this, "Added student " + s.getId() + " to the database");
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception adding student (" + s.getId() + ")", e);
            return false;
        }
        return true;
    }

    /**
     * Add a student to the student cache
     * @param Student to add
     */
    private void addStudentToCache(Student student) {
        String key = student.getId();

        if(key != null) {
            studentCache.put(new Element(key, student));
        } else {
            System.err.println("Could not add student to cache");
            Logging.error(this, "Could not add student to cache");
        }
    }

    /**
     * Retrieves a student's information from the database.
     * If there is no information about this student in the database,
     * null is returned.
     * @param studentId
     */
    private Student getStudentFromDatabase(String studentId, Connection con) {
        Student stu = null;
        try {
            String query = "SELECT name, password, email, lang, defcourse, studentNumber FROM users WHERE id=?";
            PreparedStatement st = con.prepareStatement(query);
            st.setString(1, studentId);
            ResultSet rs = st.executeQuery();

            if (!rs.next() || rs.getString(1) == null) {
                // if student not in db, return null
                return null;
            }

            stu = new Student(studentId);

            List courses = getCoursesForStudent(studentId, Student.DEFAULT_ACCESS, con);
            stu.setCourseList(courses);
            List inactive = getCoursesForStudent(studentId, UNREGISTERED, con);
            stu.setInactiveCourses(inactive);
            setResultsForStudent(stu, con);
            stu.setName(rs.getString("name"));
            stu.setPassword(rs.getString("password"));
            stu.setEmail(rs.getString("email"));
            stu.setLanguage(rs.getString("lang"));
            stu.setStudentNumber(rs.getString("studentNumber"));

            int defCourse = rs.getInt("defcourse");
            if (defCourse != 0) {
                Course c = getCourse(defCourse);
                if (c != null && stu.isOnCourse(c)) {
                    stu.setDefaultCourse(c);
                }
            }

            // Personal deadlines
            query = "SELECT rid, deadline FROM personaldeadline WHERE userid=?";
            st = con.prepareStatement(query);
            st.setString(1, studentId);
            rs = st.executeQuery();

            while(rs.next()) {
                int rid = rs.getInt("rid");
                java.util.Date d = rs.getTimestamp("deadline");
                stu.addDeadLine(rid,d);
            }

            return stu;
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting student ("+studentId+" from db", e);
        }
        return null;
    }

  
    /**
     * Retrieves a student's information that was aquired through shibboleth from the database.
     * If there is no information about this student in the database,
     * null is returned.
     * @param studentId
     */
    public StudentShibInfo getStudentShibInfo(String studentId) {
        Connection con = getDatabaseConnection();        
        
        try {
            String query = "SELECT * FROM userShibInfo WHERE id=?";
            PreparedStatement st = con.prepareStatement(query);
            st.setString(1, studentId);
            ResultSet rs = st.executeQuery();

            if (!rs.next() || rs.getString(1) == null) {
                return null;
            }

            StudentShibInfo shibInfo = new StudentShibInfo(studentId);

            shibInfo.setGender(rs.getString("gender"));
            shibInfo.setOrganization(rs.getString("organization"));
            shibInfo.setOrganizationName(rs.getString("organizationName"));
            shibInfo.setStudentNumber(rs.getString("studentNumber"));
            shibInfo.setCommonName(rs.getString("commonname"));
            shibInfo.setDisplayName(rs.getString("displayname"));
            shibInfo.setGivenName(rs.getString("givenName"));
            shibInfo.setSurname(rs.getString("surname"));
            shibInfo.setEmail(rs.getString("email"));
            shibInfo.setLang(rs.getString("lang"));

            return shibInfo;
        } catch(SQLException e) {
            Logging.error(this, "Error when getting studentShibInfo for "+studentId, e);
            e.printStackTrace();
        }
        
        return null;
    }
    
    public void saveStudentShibInfo(StudentShibInfo info) {
        if (info == null) {
            Logging.warn(this, "Calling saveStudentShibInfo with null parameter");
            return;
        }
        
        try {
            Connection con = getDatabaseConnection();
            String ps = "INSERT INTO userShibInfo(id, gender, organization, organizationName, studentNumber, commonname,"
                +  " displayname, givenname, surname, email, lang)"
                + " VALUES (?,?,?,?,?,?,?,?,?,?,?)";
            PreparedStatement st = con.prepareStatement(ps);
            st.setString(1, info.getId());
            st.setString(2, info.getGender());
            st.setString(3, info.getOrganization());
            st.setString(4, info.getOrganizationName());
            st.setString(5, info.getStudentNumber());
            st.setString(6, info.getCommonName());
            st.setString(7, info.getDisplayName());
            st.setString(8, info.getGivenName());
            st.setString(9, info.getSurname());
            st.setString(10, info.getEmail());
            st.setString(11, info.getLang());
            st.executeUpdate();

            closeDatabaseConnection(con);
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception adding studentShibInfo (" +info.getId() + ")", e);
        }
    }
    
    public void updateStudentShibInfo(StudentShibInfo info) {
        if (info == null) {
            Logging.warn(this, "Calling updateStudentShibInfo with null parameter");
            return;
        }
        
        try {
            Connection con = getDatabaseConnection();
            String ps = "UPDATE userShibInfo SET gender=?, organization=?, organizationName=?,"
                + " studentNumber=?, commonname=?, displayname=?, givenname=?, surname=?, email=?, lang=?"
                + " WHERE id=?";
            PreparedStatement st = con.prepareStatement(ps);
            st.setString(1, info.getGender());
            st.setString(2, info.getOrganization());
            st.setString(3, info.getOrganizationName());
            st.setString(4, info.getStudentNumber());
            st.setString(5, info.getCommonName());
            st.setString(6, info.getDisplayName());
            st.setString(7, info.getGivenName());
            st.setString(8, info.getSurname());
            st.setString(9, info.getEmail());
            st.setString(10, info.getLang());
            st.setString(11, info.getId());
            st.executeUpdate();

            closeDatabaseConnection(con);
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception updating studentShibInfo (" +info.getId() + ")", e);
        }
    }
    
    /*
    public Iterator getCourseIterator() {
        //return courses.iterator();
        // Are these necessary? -TA
        return null;
    }

    public Iterator getRoundIterator() {
        //return roundTemplates.iterator();
        return null;
    }


    public Iterator getExerciseIterator() {
        // XXX
        return null;
    }
    */

    /**
     * Get a course selection model containing all courses
     * @return course selection model
     */
    public CourseSelectionModel getCourseSelectionModel() {
        Connection con = getDatabaseConnection();
        List courses = getOpenCourses(con);
        closeDatabaseConnection(con);

        // 2005 EXPERIMENT HACK BEGINS //
        // Can this be removed? -TA 07/2007
        courses.remove(getCourse(0));
        courses.remove(getCourse(999));
        // 2005 EXPERIMENT HACK ENDS //


        return new CourseSelectionModel(courses);
    }

    /**
     * Get a localized version of the course selection model in the given locale.
     * @param locale
     * @return course selection model
     */
    public CourseSelectionModel getLocalizedCourseSelectionModel(String locale) {
        Connection con = getDatabaseConnection();
        List courses = getOpenCourses(con);
        closeDatabaseConnection(con);

        // 2005 EXPERIMENT HACK BEGINS //
        courses.remove(getCourse(0));
        courses.remove(getCourse(999));
        // 2005 EXPERIMENT HACK ENDS //

        return new CourseSelectionModel(courses, locale);
    }

    /**
     * Get a locale selection model for the current course
     * @return locale selection model
     */
    public LocaleSelectionModel getLocaleSelectionModel() {
        Connection con = getDatabaseConnection();
        List locales = getLocales(con);
        closeDatabaseConnection(con);
        return new LocaleSelectionModel(locales);
    }

    /**
     * Get a list of locales present in the database
     * 
     * @return list of locales
     */
    public List getLocales() {
        Connection con = getDatabaseConnection();
        List locales = getLocales(con);
        closeDatabaseConnection(con);
        return locales;
    }

    /*
    public  TemplateSelectionModel getTemplateSelectionModel() {
        return null;
    }
    */

    public synchronized int setResult(String studentId,
                        String courseId, int round,
                        int exercise, int points) {
        return setResult(studentId, courseId, round, exercise, points, null, OK);
    }

    public synchronized int setResult(String studentId,
                        String courseId, int round,
                        int exercise, int points, int status) {
        return setResult(studentId, courseId, round, exercise, points, null, status);
    }

    /**
     * Add a submission.
     * @param studentId
     * @param courseId
     * @param round
     * @param exercise
     * @param points
     * @param fileName
     * @param status
     * @return
     */
    public synchronized int setResult(String studentId, String courseId, int round, int exercise,
                                      int points, String fileName, int status)
    {
        Student st = getStudent(studentId);

        if(st == null) {
            return STUDENT_NOT_FOUND;
        }

        int i = Integer.parseInt(courseId);
        // For results coming to non-existent courses we just silently ignore them.
        if (i == ALL_COURSES || i == DUMMY_COURSE) {
            return OK;
        }

        Course c = getCourse(i);
        if(c == null) {
            return COURSE_NOT_FOUND;
        } else if (!st.isOnCourse(c)) {
            return STUDENT_NOT_ON_COURSE;
        }

        boolean dbsuccess = addResult(studentId, courseId, round, exercise,
                        points, fileName, status);
        if (!dbsuccess) {
            return DATABASE_ERROR;
        }
        if (status == OK) {
            st.setResult(c,round,exercise,points);
        } else if (status == AFTER_DEADLINE) {
            Round r = (Round)DataStorage.getInstance().getCourse(Integer.parseInt(courseId)).getRoundList().get(round-1);
            int makeupPoints = (int)Math.ceil(points*r.getMakeupFactor());
            st.setResult(c,round,exercise,makeupPoints);
        }
        return OK;
    }

    /**
     * Register the student with the given userid to the given course
     * @param studentId
     * @param courseId
     */
    public void registerStudentToCourse(String studentId, int courseId) {
        String query = "SELECT userpriviledge FROM registrations WHERE userid=? AND cid=" + courseId +
            " AND userpriviledge<" + INSTRUCTOR;

        try {
            Connection con = getDatabaseConnection();

            PreparedStatement st = con.prepareStatement(query);
            st.setString(1, studentId);
            ResultSet rs = st.executeQuery();

            if(rs.next()) {
                int priviledge = rs.getInt("userpriviledge");
                if(priviledge == STUDENT) {
                    // already registered, do nothing
                    return;
                }
                query = "UPDATE registrations set userpriviledge = " + STUDENT + " WHERE userid=?"+
                    " AND cid=" + courseId + " AND userpriviledge="+priviledge;
            } else {
                query = "INSERT INTO Registrations VALUES (?, '"+courseId+"')";
            }

            st = con.prepareStatement(query);
            st.setString(1, studentId);
            int i = st.executeUpdate();
            if(i != 1) {
                Logging.error(this, "Error in updating registrations with command: " + query + "\nCancelling actions");
                st.cancel();
                // Now we should give some kind of error feedback. XXX TODO
            }
            closeDatabaseConnection(con);
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Could not register student ("+studentId+") to course "+courseId);
            Logging.error(this, "Query: " + query);
            Logging.error(this, e.toString());
        }
    }

    /**
     * Unregister a student from a course
     * @param studentId
     * @param courseId
     */
    public void unregisterStudentFromCourse(String studentId, int courseId) {
        try {
            Connection con = getDatabaseConnection();
            String update = "UPDATE registrations set userpriviledge = "
                    + UNREGISTERED + " WHERE userid=?"
                    + " AND cid=" + courseId + " AND userpriviledge="
                    + STUDENT;
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, studentId);
            int i = st.executeUpdate();
            if (i != 1) {
                Logging.error(this,
                              "Error in updating registrations with command: "
                                      + update + "\nCancelling actions");
                st.cancel();
                // Now we should give some kind of error feedback. XXX TODO
            }
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Could not unregister student ("
                    + studentId + ") to ccourse " + courseId);
        }
    }

    /**
     * Get a round containing all exercises in the db
     * @return Round containing all exercises
     */
    public Round getAllExercises() {
        Connection con = getDatabaseConnection();
        Round r = getRoundFromDatabase(0, con);
        closeDatabaseConnection(con);
        return r;
    }


    /**
     * Get the submissions for a certain course, round, exercise, student combination
     * @param cid
     * @param rid
     * @param enumber
     * @param studentId
     * @return
     */
    public List getAnswers(int cid, int rid, int enumber, String studentId) {
        Connection con = null;
        try {
            con = getDatabaseConnection();
            Statement st = con.createStatement();
            String query = "SELECT eid FROM exercises WHERE rid="+rid+" AND number="+enumber;
            ResultSet rs = st.executeQuery(query);
            if (rs.next()) {
                int eid = rs.getInt("eid");
                query = "SELECT points, tstamp,status,filename FROM results WHERE cid="+cid+" AND rid="+rid+" AND eid="+eid+" AND "+
                    "userid=? ORDER BY tstamp";// AND status <= " + AFTER_DEADLINE;
                PreparedStatement pst = con.prepareStatement(query);
                pst.setString(1, studentId);
                rs = pst.executeQuery();
                List l = new LinkedList();

                while(rs.next()) {
                    int points = rs.getInt(1);
                    java.util.Date d = rs.getTimestamp(2);
                    int status = rs.getInt(3);
                    String filename = rs.getString(4);
                    l.add(new Answer(points, status,d,filename));
                }

                closeDatabaseConnection(con);
                return l;
            }
            return null;
        } catch(Exception e) {
            e.printStackTrace();
            Logging.error(this, "Could not fetch answer to exercise " + enumber + " on round with id " + rid + " for student "
                + studentId);
            closeDatabaseConnection(con);
        }
        return null;
    }

    /**
     * Add a personal deadline for a student
     * @param s
     * @param id
     * @param time
     * @param username
     * @return
     */
    public boolean addPersonalDeadline(Student s, int rid, String time, String username) {
        Connection con = null;
        try {
            con = getDatabaseConnection();
            String query = "SELECT * FROM personaldeadline WHERE userid=? AND rid="+rid;
            PreparedStatement st = con.prepareStatement(query);
            st.setString(1, s.getId());
            ResultSet rs = st.executeQuery();
            String update = "";
            if(rs.next()) {
                update = "UPDATE personaldeadline SET deadline=?, grantedby=? WHERE "+
                "userid=? AND rid="+rid;
                st = con.prepareStatement(update);
                st.setString(1, time);
                st.setString(2, username);
                st.setString(3, s.getId());
            } else {
                update ="INSERT INTO personaldeadline(userid, rid, deadline, grantedby) VALUES(?,"+rid+",?,?)";
                st = con.prepareStatement(update);
                st.setString(1, s.getId());
                st.setString(2, time);
                st.setString(3, username);
            }
            int ret = st.executeUpdate();
            closeDatabaseConnection(con);
            if(ret < 1) {
                return false;
            }
            return true;

        } catch(Exception e) {
            e.printStackTrace();
            Logging.error(this, "Could not set personal deadline for student " + s.getId());
            closeDatabaseConnection(con);
        }
        return false;
    }


    /**
     * Change a exercise description in the database
     * @param descr
     * @param locale
     * @param eid
     * @return
     */
    public boolean updateDescription(String descr, String locale, int eid) {
        if(descr.indexOf('\'') != -1) {
            StringBuffer buf = new StringBuffer(descr);
            int index = 0;
            while(index < buf.length() && buf.indexOf("\'", index) != -1) {
                index = buf.indexOf("\'", index);
                buf = buf.insert(index,'\'');
                index = index+2;
            }
            descr = buf.toString();
        }
        Connection con = getDatabaseConnection();
        boolean ret = false;
        try {
            String update = "UPDATE exercisedescription SET description=? WHERE lang=? AND eid="+eid;
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, descr);
            st.setString(2, locale);
            int val = st.executeUpdate();
            if(val < 1) {
                Logging.error(this, "update to eid " + eid+ " in lang " + locale +" had no effect.");
            } else {
                ret = true;
            }
        } catch(Exception e) {
            e.printStackTrace();
            ret = false;
        }
        closeDatabaseConnection(con);
        return ret;
    }

    /**
     * Change a course description in the database
     * @param description
     * @param locale
     * @param cid
     * @return
     */
    public boolean updateCourseDescription(String description, String locale, int cid) {
        if(description.indexOf('\'') != -1) {
            StringBuffer buf = new StringBuffer(description);
            int index = 0;
            while(index < buf.length() && buf.indexOf("\'", index) != -1) {
            index = buf.indexOf("\'", index);
            buf = buf.insert(index,'\'');
            index = index+2;
            }
            description = buf.toString();
        }
        Connection con = getDatabaseConnection();
        boolean ret = false;
        try {
            String update = "UPDATE coursedescription SET description=? WHERE lang=? AND cid="+cid;
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, description);
            st.setString(2, locale);
            int val = st.executeUpdate();
            if(val < 1) {
            Logging.error(this, "updateCourseDescription: update to cid " + cid+ " in lang " + locale +" had no effect.");
            } else {
            ret = true;
            }
        } catch(Exception e) {
            e.printStackTrace();
            ret = false;
        }
        closeDatabaseConnection(con);
        return ret;
    }

    /**
     * Add a course description (in a new locale) to the given course
     * @param name
     * @param description
     * @param locale
     * @param cid
     * @return
     */
    public synchronized boolean addCourseData(String name, String description, String locale, int cid) {
        if(description.indexOf('\'') != -1) {
            StringBuffer buf = new StringBuffer(description);
            int index = 0;
            while(index < buf.length() && buf.indexOf("\'", index) != -1) {
            index = buf.indexOf("\'", index);
            buf = buf.insert(index,'\'');
            index = index+2;
            }
            description = buf.toString();
        }
        if(description == null)
            description = "";
        if(getCourse(cid).getSupportedLocales().contains(locale)) {
            Logging.error(this, "addCourseData: course " + cid + " already has localized data in locale " + locale + ".");
            return false;
        }
        if(name == null || locale == null) {
            Logging.error(this, "addCourseData: name or locale has a null value. Name= '"+ name +"' locale= '" + locale +"'.");
            return false;            
        }
        Connection con = getDatabaseConnection();
        boolean ret = false;
        try {
            String update = "INSERT INTO coursedescription VALUES(" + cid + ", ?, ?, ?)";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, name);
            st.setString(2, locale);
            st.setString(3, description);
            int val = st.executeUpdate();
            if(val < 1) {
            Logging.error(this, "addCourseData: addition to cid " + cid+ " in lang " + locale +" had no effect.");
            } else {
            ret = true;
            }
        }catch(Exception e) {
            e.printStackTrace();
            ret = false;
        }
        closeDatabaseConnection(con);
        return ret;
    }

    /**
     * Add a description (in a new locale) to the given exercise
     * 
     * @param name
     * @param description
     * @param locale
     * @param eid
     * @return
     */
    public synchronized boolean addExerciseData(String name, String description, String locale, int eid) {
        if(description.indexOf('\'') != -1) {
            StringBuffer buf = new StringBuffer(description);
            int index = 0;
            while(index < buf.length() && buf.indexOf("\'", index) != -1) {
                index = buf.indexOf("\'", index);
                buf = buf.insert(index,'\'');
                index = index+2;
            }
            description = buf.toString();
        }
        if(description == null)
            description = "";
        if(getExerciseTemplate(eid) != null && getExerciseTemplate(eid).getSupportedLocales().contains(locale) ) {
            Logging.error(this, "addExerciseData: exercise " + eid + " already has localized data in locale " + locale + ".");
            return false;
        }
        if(name == null || locale == null) {
            Logging.error(this, "addExerciseData: name or locale has a null value. Name= '"+ name +"' locale= '" + locale +"'.");
            return false;            
        }
        Connection con = getDatabaseConnection();
        boolean ret = false;
        try {
            String update = "INSERT INTO exercisedescription VALUES(?, ?, ?, " + eid + ")";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, description);
            st.setString(2, name);
            st.setString(3, locale);            
            int val = st.executeUpdate();
            if(val < 1) {
                Logging.error(this, "addExerciseData: addition to eid " + eid + " in lang " + locale +" had no effect.");
            } else {
                ret = true;
            }
        }catch(Exception e) {
            e.printStackTrace();
            ret = false;
        }
        closeDatabaseConnection(con);
        return ret;
    }

    /**
     * Add a round description (in a new locale) to the given round
     * @param name
     * @param description
     * @param locale
     * @param rid
     * @return
     */
    public synchronized boolean addRoundData(String name, String description, String locale, int rid) {
        if(description.indexOf('\'') != -1) {
            StringBuffer buf = new StringBuffer(description);
            int index = 0;
            while(index < buf.length() && buf.indexOf("\'", index) != -1) {
            index = buf.indexOf("\'", index);
            buf = buf.insert(index,'\'');
            index = index+2;
            }
            description = buf.toString();
        }
        if(description == null)
            description = "";
        Connection con = getDatabaseConnection();
        Round r = getRoundFromDatabase(rid, con);
        closeDatabaseConnection(con);
        if(r.getSupportedLocales().contains(locale)) {
            Logging.error(this, "addRoundData: round " + rid + " already has localized data in locale " + locale + ".");
            return false;
        }
        if(locale == null) {
            Logging.error(this, "addRoundData: locale has a null value. Locale= '" + locale +"'.");
            return false;
        }
        con = getDatabaseConnection();
        boolean ret = false;
        try {
            String update = "INSERT INTO rounddescription VALUES(" + rid + ", ?, ?, ?)";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, name);
            st.setString(2, locale);
            st.setString(3, description);
            int val = st.executeUpdate();
            if(val < 1) {
            Logging.error(this, "addRoundData: addition to rid " + rid+ " in lang " + locale +" had no effect.");
            } else {
            ret = true;
            }
        } catch(Exception e) {
            e.printStackTrace();
            ret = false;
        }
        closeDatabaseConnection(con);
        return ret;
    }

    /**
     * Update the round description for the given round
     * @param description
     * @param locale
     * @param rid
     * @return
     */
    public boolean updateRoundDescription(String description, String locale, int rid) {
        if(description.indexOf('\'') != -1) {
            StringBuffer buf = new StringBuffer(description);
            int index = 0;
            while(index < buf.length() && buf.indexOf("\'", index) != -1) {
            index = buf.indexOf("\'", index);
            buf = buf.insert(index,'\'');
            index = index+2;
            }
            description = buf.toString();
        }
        Connection con = getDatabaseConnection();
        boolean ret = false;
        try {
            String update = "UPDATE rounddescription SET description=? WHERE lang=? AND rid="+rid;
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, description);
            st.setString(2, locale);
            int val = st.executeUpdate();
            if(val < 1) {
            Logging.error(this, "updateRoundDescription: update to rid " + rid+ " in lang " + locale +" had no effect.");
            } else {
            ret = true;
            }
        } catch(Exception e) {
            e.printStackTrace();
            ret = false;
        }
        closeDatabaseConnection(con);
        return ret;
    }

    /**
     * Get the date of the first submission for a course
     * @param course
     * @return Date
     */
    public java.util.Date getDateOfFirstAnswer(Course c) {
        Connection con = getDatabaseConnection();
        try {
            String query = "SELECT min(tstamp) FROM results WHERE cid="+c.getId();
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(query);
            if(rs.next()) {
            java.util.Date d = rs.getTimestamp(1);
            closeDatabaseConnection(con);
            return d;
            }
            closeDatabaseConnection(con);
            return null;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }


    /**
     * Gets the activity of students on the given course between the given
     * start and end dates.
     * @param c course
     * @param start start date
     * @param end end date
     * @return student activity by hour
     */
    public int[] getStudentActivityByHour(Course c, java.util.Date start, java.util.Date end) {
        return getStudentActivity(c,start,end,MILLIS_IN_HOUR);
    }

    /**
     * Gets the student activity by day on a given course
     * @param c course
     * @param start start date
     * @param end end date
     * @return student activity by day
     */
    public int[] getStudentActivityByDay(Course c, java.util.Date start, java.util.Date end) {
        return getStudentActivity(c,start,end,MILLIS_IN_HOUR*24);
    }

    /**
     * Gets the average student activity by time of day on a given course
     * 
     * @param c course
     * @return average student activity
     */
    public int[] getStudentActivityAverage(Course c) {
        GregorianCalendar cal = new GregorianCalendar();
        java.util.Date d = c.getOpen();
        if(d == null) {
            d = getDateOfFirstAnswer(c);
        }
        if(d == null) {
            return null;
        }

        cal.setTime(d);
        cal.set(Calendar.HOUR, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        d = new java.util.Date();

        return getStudentActivity(c,cal.getTime(),d,MILLIS_IN_HOUR);
    }

    /**
     * Gets student activity for a specified time of a given course.
     * Divisor is assumed to be a multiple of MILLIS_IN_HOUR. It tells how many hours are grouped together.
     * @param c course
     * @param start start date
     * @param end end date
     * @param divisor divisor
     * @return student activity data
     */
    public int[] getStudentActivity(Course c, java.util.Date start, java.util.Date end, double divisor) {
	Connection con = getDatabaseConnection();
	try {
	    // tells how many hours there are in a single block.
	    int hourDivisor = (int)(divisor / MILLIS_IN_HOUR);
	    int[] activity = new int[nroHours(start,end)/hourDivisor];
	    if(activity.length > 0) {
		ResultSet rs = getResultsByDateFromDatabase(start,end,c.getId(),con);
		GregorianCalendar gc = new GregorianCalendar();
		//SimpleDateFormat tf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		gc.setTime(start);
		setToHour(gc);
		long time = gc.getTimeInMillis();
		java.util.Date d = null;
		while(rs.next()) {
		    d = rs.getTimestamp(2);
		    activity[(int)Math.floor((d.getTime()-time)/divisor)]++;
		}
		rs.close();
	    }
	    closeDatabaseConnection(con);
	    return activity;
	}catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Error while fetching student activity results");
	}
	return null;
    }


    /**
     * Moves the parameter student from one course to another. If the
     * courses share exercises, all answers are moved from old course
     * to new one.  Exercises on the courses are considered identical,
     * if they have the same exercise id and point maximum. Exercises
     * that cannot be moved are left to the old course. 
     *
     * @param s the student to be moved
     * @param from the Course the Student is removed from
     * @param to the Course the Student is added to
     */
    public boolean moveStudent(Student s, Course from, Course to) {
	if(s == null || from == null || to == null) {
	    Logging.error(this, "Cannot move student " + s + " from course " + from + " to course " + to +".");
	    return false;
	}
	registerStudentToCourse(s.getId(), to.getId());
	try {
	    Connection con = getDatabaseConnection();
	    PreparedStatement st;// = con.createStatement();
	    Round oldr, newr;
	    Exercise olde, newe;
	    List oldRounds, newRounds, oldExercises, newExercises;
	    int oldrid, newrid, eid, success;
	    String update;
	    oldRounds = from.getRoundList();
	    newRounds = to.getRoundList();
	    // this horrible hack of four inner loops assigns correct
	    // values to new and old Rounds and Exercises
	    // in order to move the exercises
	    for(int i = 0; i < oldRounds.size();i++) {
		oldr = (Round)oldRounds.get(i);
		oldExercises = oldr.getExerciseList();
		for(int j = 0; j < oldExercises.size();j++) {
		    olde = (Exercise)oldExercises.get(j);
		    oldrid = newrid = eid = -1;
		loop: 
		    for(int k = 0; k < newRounds.size(); k++) {
			newr = (Round)newRounds.get(k);
			newExercises = newr.getExerciseList();
			for(int l = 0; l < newExercises.size(); l++) {
			    newe = (Exercise) newExercises.get(l);
			    // Now, if olde and newe have same id and the same maximum number of points,
			    // we consider them to be the same exercise.
			    if(newe.getMax() == olde.getMax() && 
			       newe.getExerciseTemplate().getId() == olde.getExerciseTemplate().getId()) {
				oldrid = oldr.getId();
				newrid = newr.getId();
				eid = olde.getExerciseTemplate().getId();
				break loop;
			    }
			}
		    }
		    if(oldrid == -1) {
			continue;
		    }

		    // At this point the values of the old and new exercises have been assigned
		    // now we can modify the results table.
		    update = "UPDATE results SET cid=" + to.getId() + ", rid=" + newrid + " WHERE userid=?"
			    + " AND cid=" + from.getId() + " AND rid=" + oldrid + " AND eid=" + eid;
            st = con.prepareStatement(update);
            st.setString(1, s.getId());
		    st.executeUpdate();
		}
	    }
	    update = "UPDATE registrations SET userpriviledge="+MOVED+" WHERE userid=?" +
		     " AND cid=" + from.getId();
        st = con.prepareStatement(update);
        st.setString(1, s.getId());
	    success = st.executeUpdate();
	    closeDatabaseConnection(con);
	    if(success < 1) {
		Logging.error(this, "Could not move student " + s + " from course " + from + ": DB error during update");
		return false;
	    }
	}catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Error while moving a student's answers from course " + from + " to course " + to + ".");
	    return false;
	}
	return true;
    }


    /**
     * Adds a lock for one exercise in the system. This lock is used by the
     * monitor interface to prevent two instructors from modifying the same
     * exercise simultaneously. The key must be the exercise id, while the
     * id is the id of the <strong>lock</strong> and must be unique for each
     * lock. Using the Visit object of the user is OK.
     *
     * @param eid Key for this lock, the id of the exercise to be modified
     * @param id the id of the lock, must be unique in the system.
     */ 
    public synchronized Lock addLock(Integer eid, Object id) {
	Lock l = (Lock)locks.get(eid);
	long time = Calendar.getInstance().getTimeInMillis();
	if(l != null && (time - l.getTimeStamp()) < LOCK_TIMEOUT && l.getId() != id) {
	    // the lock has already created, return null
	    // to indicate that lock cannot be created
	    return null;
	}
	if(l != null) {
	    locks.remove(eid);
	}
	l = new Lock(eid, id, time);
	locks.put(eid, l);
	return l;
    }
    
    /**
     * Remove the lock for one exercise in the system. Other instructors are now
     * allowed to modify the exercise. 
     * @param eid Key for this lock, the id of the exercise to be modified
     * @return lock the removed lock
     */
    public synchronized Lock removeLock(Integer eid) {
        return (Lock)locks.remove(eid);
    }

    /**
     * Renew the lock for one exercise in the system.
     * 
     * @param eid Key for this lock, the id of the exercise to be modified
     * @param id the id of the lock, must be unique in the system.
     * @return
     */
    public synchronized Lock renewLock(Integer eid, Object id) {
	Lock l = (Lock)locks.get(eid);
	if(l != null && l.getId() == id) { // yes, using == here is intentional
	    long time = Calendar.getInstance().getTimeInMillis();
	    if((time - l.getTimeStamp()) < LOCK_TIMEOUT) {
		l.setTimeStamp(time);
		return l;
	    } else { // lock is old and should therefore be removed.
		removeLock(eid);
	    }
	}
	return null;
    }
    
    public synchronized Lock getLock(Integer eid) {
  	return (Lock)locks.get(eid);
    }
    
    private GregorianCalendar setToHour(GregorianCalendar gc) {
	gc.set(Calendar.MINUTE, 0);
	gc.set(Calendar.SECOND, 0);
	gc.set(Calendar.MILLISECOND,0);
	return gc;
    }

    /**
     * Calculates how many hours there are between dates d1 and
     * d2. The comparison does not use the actual times to calculate the difference,
     * but calculates how many hours are between start of the hour given by d1 and the
     * end of the hour given by d2. 
     *
     * @return the number of hours between the two dates, or 0 if d1 is not before d2
     */
    private int nroHours(java.util.Date d1, java.util.Date d2) {
	GregorianCalendar c1 = new GregorianCalendar();
	c1.setTime(d1);
	GregorianCalendar c2 = new GregorianCalendar();
	c2.setTime(d2);
	if(c1.after(c2)) 
	    return 0;
	setToHour(c1);
	if(c2.get(Calendar.MINUTE) > 0 || c2.get(Calendar.SECOND) > 0 || c2.get(Calendar.MILLISECOND) > 0) {
	    c2.setLenient(true);
	    c2.set(Calendar.HOUR_OF_DAY, c2.get(Calendar.HOUR_OF_DAY) +1);
	    setToHour(c2);
	}
	double diff = c1.getTimeInMillis();
	diff = c2.getTimeInMillis() - diff;
	return (int)Math.ceil(diff/MILLIS_IN_HOUR);
    }

    // --- XXX Private methods XXX ---


    /**
     * Tells if the given course is found in the course cache and that the cache element
     * is available in memory.
     * 
     * @param c
     * @return
     */
    private boolean containsCourse(Course c) {
        // everything should be stored in memory, so need just to check that
        return courseCache.isElementInMemory("" + c.getId());
    }

    /**
     * Add a course to the course cache
     * 
     * @param course
     * @param id
     */
    private void addCourse(Course c, int id) {
        courseCache.put(new Element("" + id, c));
    }

    /**
     * Get a course from cache
     *
     * @param id
     * @return
     */
    private Course getCourseFromCache(int id) {
        Element e = courseCache.get("" + id);
        Course c = null;
        if (e != null) {
            c = (Course)e.getValue();
        }
        if(c == null || c.getId() != id) {
            return null;
        }

        return c;
    }

    private void removeCourse(Course c) {
	// May someday be activated. Now it's just too bloody difficult!
//  	int i = courses.indexOf(c);
//  	if(i != -1) {
//  	    c.removeAllRounds();
//  	    courses.remove(i);
//  	}
    }


    /**
     * Adds an exercise template to the cache
     * 
     * @param t exercise template
     * @param id template id
     */
    private void addExerciseTemplate(ExerciseTemplate t, int id) {
        exerciseTemplateCache.put(new Element("" + id, t));
    }

//  public synchronized void removeExerciseTemplate(ExerciseTemplate t) {
//      if(exerciseTemplates.indexOf(t) != -1 && t.getUses() < 1) {
//          exerciseTemplates.remove(t);
//      }
//  }

    private boolean containsExerciseTemplate(ExerciseTemplate e) {
        return exerciseTemplateCache.isElementInMemory("" + e.getId());
    }

    /**
     * Gets a exercise template from the cache.
     *
     * @param id
     * @return
     */
    private ExerciseTemplate getExerciseTemplate(int id) {
        Element e = exerciseTemplateCache.get("" + id);
        ExerciseTemplate ex = null;
        if (e != null) {
            ex = (ExerciseTemplate) e.getValue();
        }

        if(ex == null || ex.getId() != id) {
            return null;
        }

        return ex;
    }

    /**
     * Handles the DataStorage initialization.
     */
    private void initialize() {
        //initializeCourse();
        //initializeStudent();
        loadProperties();

        startXmlRpc();

        initializeCaches();
    }


    // ---- XXX initialization Stuff XXX ----

    /**
     * Starts the xml-rpc server.
     */
    private void startXmlRpc() {
        int port = Integer.parseInt(System.getProperty("xmlrpc.port"));

        try {
            WebServer server = new WebServer(port);
            server.addHandler("result", new ResultHandler());
            //server.setParanoid(true);
            //server.acceptClient("127.0.0.1");
            Logging.info(this, "XML-RPC Server started in port "+port+".");
            server.start();
        } catch (Exception e) {
            Logging.error(this, "XML-RPC failed to start. " + e);
            return;
        }
    }

    /**
     * Initializes the caches.
     */
    private void initializeCaches() {
        CacheManager.create();
        CacheManager manager = CacheManager.getInstance();
        courseCache = manager.getCache(System.getProperty("course.cache", "courses"));
        exerciseTemplateCache = manager.getCache(System.getProperty("exercise.cache", "templates"));
        studentCache = manager.getCache(System.getProperty("student.cache","students"));
    }

    /**
     * Loads server properties from the <strong>DataStorage.properties</strong> file.
     */
    private void loadProperties() {
        InputStream properties = this.getClass().getResourceAsStream("DataStorage.properties");
        try {
            System.getProperties().load(properties);
        } catch (IOException e) {
            System.err.println("Properties could not be loaded for DataStrorage");
            e.printStackTrace();
            Logging.error(this, "Properties could not be loaded for DataStrorage", e);
        }
    }

    // ---- XXX Database methods start here XXX ----


    /**
     * Closes the given database connection
     * 
     * @param c database connection
     */
    private void closeDatabaseConnection(Connection c) {
        if(c == null) {
            return;
        }

        try {
            c.close();
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception closing db connection", e);
        }
    }

    /**
     * Builds the results for all students on a single course. Used by instructor UI.
     */
    private List getStudentResultsForCourseFromDatabase(Course course, Connection con) {
    	try {
        	// fetch makeup factors and round numbers for each round
            String query = "SELECT id, number, makeupfactor FROM round WHERE cid="+course.getId() + "ORDER BY number";
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(query);
            // store makeup factors
            HashMap makeup = new HashMap();
            // store round numbers
            HashMap roundNros = new HashMap();
            // store exercise numbers
            HashMap exerciseNros = new HashMap();
            Integer rid, eid, rnro, enro;
            // int points, subs, status;
            Float factor;

            Logging.debug("Starting student retrieval from DB");
            Logging.debug("Fetch exercise numbers from Database");
            while(rs.next()) {
                rid = new Integer(rs.getInt("id"));
                rnro = new Integer(rs.getInt("number"));
                factor = new Float(rs.getFloat("makeupfactor"));
                makeup.put(rid,factor);
                roundNros.put(rid,rnro);
                query = "SELECT eid, number FROM exercises where rid = " + rid.intValue();
                st = con.createStatement();
                ResultSet drs = st.executeQuery(query);
                // XXX This does not work if the same exercise is on several rounds!
                while(drs.next()) {
                    eid = new Integer(drs.getInt("eid"));
                    enro = new Integer(drs.getInt("number"));
                    exerciseNros.put(eid,enro);
                }
            }
            rs.close();
            Logging.debug("Numbers fetched, getting students");
            // fetch all students that are registered to the course, ordered by userid
            query = "SELECT userid, userpriviledge, name, email FROM registrations, users WHERE cid=" + course.getId() + " AND (userpriviledge = " + STUDENT + 		" OR userpriviledge = " + UNREGISTERED + ") AND userid = id ORDER BY userid";
            ResultSet students = st.executeQuery(query);
            Logging.debug("Students fetched, fetching answers");
            // fetch student results, again ordered by userid
            // XXX HOW TO HANDLE MAKE-UP SUBMISSIONS CORRECTLY?
            // Erik: changed the query from grouping by status to only showing the bigger status flag, MAX (status), to cope with multiple allowed status flags
            query =  "SELECT userid, MAX(points) AS points, eid, rid, COUNT(points) AS subs, MAX(status) AS status FROM results WHERE cid=" +
            course.getId() + " AND (status =" + OK + " OR status=" + OK_MODIFIED + " ) GROUP BY userid, rid, eid ORDER BY userid";
            // added parameters for postgresql 8.x compatibility
            st = con.createStatement(ResultSet.TYPE_SCROLL_SENSITIVE, ResultSet.CONCUR_READ_ONLY);
            ResultSet answers = st.executeQuery(query);
            Logging.debug("Fetched answers");
            Logging.debug("fetching make-up -answers");
            
            // Erik: changed the query from grouping by status to only showing the bigger status flag, MAX (status), to cope with multiple allowed status flags
            query =  "SELECT userid, MAX(points) AS points, eid, rid, COUNT(points) AS subs, MAX(status) AS status FROM results WHERE cid=" + 
            course.getId() + " AND (status =" + AFTER_DEADLINE + " OR status=" + AFTER_DEADLINE_MODIFIED + ") GROUP BY userid, rid, eid ORDER BY userid";
            st = con.createStatement(ResultSet.TYPE_SCROLL_SENSITIVE, ResultSet.CONCUR_READ_ONLY);
            ResultSet makeUpAnswers = st.executeQuery(query);
            Logging.debug("Fetched make-up -answers");
            String studentId = "";
            // String studentName = "";
            String currentId = "";
            //Student stu = null;
            List results = new LinkedList();
            List courses = new LinkedList();
            courses.add(course);

            while(students.next()) {
            	studentId = students.getString("userid");
            	//Call to getStudent takes ages and does lots of unnecesary work, and therefore is not used. -jtn 12.9.07
                //Student stu = getStudent(studentId);
            	Student stu = new Student(studentId);
            	stu.setName(students.getString("name"));
            	stu.setEmail(students.getString("email"));
            	stu.setStudentNumber(studentId);
                Logging.debug("Fetching answers for student "+ studentId);

                if (students.getInt(2) == UNREGISTERED) {
                    stu.setUnregistered(true);
                }

                query = "SELECT rid, deadline FROM personaldeadline WHERE userid=?";
                PreparedStatement pst = con.prepareStatement(query);
                pst.setString(1, studentId);
                rs = pst.executeQuery();
                while(rs.next()) {
                    int r = rs.getInt("rid");
                    java.util.Date dl = rs.getTimestamp("deadline");
                    stu.addDeadLine(r,dl);
                }
                rs.close();
                
                stu.setCourseList(courses);
                stu.setCourse(course);
                results.add(stu);

                if(answers.next()) {
                    currentId = answers.getString("userid");
                    while(studentId.compareTo(currentId) > 0) {
                        if(answers.next()) {
                            currentId = answers.getString("userid");
                        } else {
                            break;
                        }
                    }

                    while(studentId.equals(currentId)) {
                        addAnswer(stu.getResults(course),answers, roundNros, exerciseNros,makeup);
                        if(answers.next()) {
                            currentId = answers.getString("userid");
                        } else {
                            break;
                        }
                    }
                    answers.previous();
                }

                if(makeUpAnswers.next()) {
                    currentId = makeUpAnswers.getString("userid");
                    while (studentId.compareTo(currentId) > 0) {
                        if(makeUpAnswers.next()) {
                            currentId = makeUpAnswers.getString("userid");
                        } else {
                            break;
                        }
                    }

                    while (studentId.equals(currentId)) {
                        addAnswer(stu.getResults(course),makeUpAnswers, roundNros, exerciseNros,makeup);
                        if(makeUpAnswers.next()) {
                            currentId = makeUpAnswers.getString("userid");
                        } else {
                            break;
                        }
                    }
                    makeUpAnswers.previous();
                }
            }
            Logging.debug("Answers fetched");
            return results;
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception while getting course results", e);
        }
        return null;
    }

    /**
     * Helper function for getStudentResultsForCourseFromDatabase
     */
    private void addAnswer(Results res, ResultSet answers, HashMap roundNros, HashMap exerciseNros, HashMap makeup) throws SQLException {
        Integer eid, rid, rnro, enro;
        int points, subs,status,oldSub;
        // String query;
        eid = rid = rnro = enro = null;
        // fetch round and exercise ID
        eid = new Integer(answers.getInt("eid"));
        rid = new Integer(answers.getInt("rid"));
        // if round ID in hashmap, use it
        if(roundNros.containsKey(rid)) {
            rnro = (Integer)roundNros.get(rid);
        } else {
            Logging.error("Could not find round number for round id " + rid);
            return;
        }

        // ditto for exercise id
        if(exerciseNros.containsKey(eid)) {
            enro = (Integer)exerciseNros.get(eid);
        } else {
            Logging.error("Could not find exercise number for exercise id " + eid);
            return;
        }
        points = answers.getInt("points");
        subs = answers.getInt("subs");
        status = answers.getInt("status");
        // what happens depends on the
        if ((status == OK || status == OK_MODIFIED) && rnro != null && enro != null) {
            res.setResults(rnro.intValue(),enro.intValue(),points,subs);
        } else if (status == AFTER_DEADLINE || status == AFTER_DEADLINE_MODIFIED) {
            int oldRes = res.getResult(rnro.intValue(),enro.intValue());
            oldSub = res.getSubmissions(rnro.intValue(),enro.intValue());
            Float factor = (Float)makeup.get(rid);
            if(factor != null) {
                points = (int)Math.ceil(points*factor.floatValue());
                if(points < oldRes) {
                    points = oldRes;
                }
            } else {
                Logging.error(this, "Could not find makeup factor for round " + rid + " from software cache");
                points = oldRes;
            }
            subs = subs+oldSub;
            res.setResults(rnro.intValue(),enro.intValue(),points,subs);
        }

        return;
    }

    public void setStudentField(String studentId, String field, String value) {
        try {
            Connection con = getDatabaseConnection();
            // make sure student is in db
            Student s = getStudentFromDatabase(studentId, con);
            String update = "UPDATE users SET " + field + "=? WHERE id=?;";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, value);
            st.setString(2, studentId);
            st.executeUpdate();
            closeDatabaseConnection(con);
            studentCache.remove(s.getId());
            // System.out.println("FIELD CHANGED");
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception setting student field ("+studentId+") " + field + "=" + value, e);
        }
    }

    /**
     * Retrieves a Monitor's information from the database. 
     */
    private Student getMonitorFromDatabase(String userName, Connection con) {
        Student stu = null;
        try {
            String query = "SELECT name, password, defcourse FROM users WHERE id=?";
            PreparedStatement st = con.prepareStatement(query);
            st.setString(1, userName);
            ResultSet rs = st.executeQuery();
            if(!rs.next()) {
                // No such user found
                return null;
            }

            stu = new Student(userName);
            List courses = getCoursesForStudent(userName, Student.INSTRUCTOR_ACCESS, con);
            if(courses.size() == 0) {
                return null;
            }

            stu.setAccessLevel(Student.INSTRUCTOR_ACCESS);
            stu.setCourseList(courses);
            stu.setName(rs.getString("name"));
            stu.setPassword(rs.getString("password"));
            int defCourse = rs.getInt("defcourse");
            if (defCourse != 0) {
                Course c = getCourse(defCourse);
                if (c != null && stu.isOnCourse(c)) {
                    stu.setDefaultCourse(c);
                }
            }
            query = "SELECT * FROM Registrations WHERE userId = ? AND cid = " + 0;
            st = con.prepareStatement(query);
            st.setString(1, userName);
            rs = st.executeQuery();
            if(rs.next()) {
                stu.setAccessLevel(rs.getInt("userpriviledge"));
            }
            return stu;
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting monitor "+userName+" from db", e);
        }
        return null;
    }

    public List searchStudents(String nameToSearch, String studentNumberToSearch, String idToSearch, List courses) {
        if (nameToSearch == null) {
            nameToSearch = "";
        }

        if (studentNumberToSearch == null) {
            studentNumberToSearch = "";
        }

        if (idToSearch == null) {
            idToSearch = "";
        }

        Connection c = getDatabaseConnection();
        List foundStudents = searchStudentsFromDatabase(idToSearch, nameToSearch, studentNumberToSearch, c);
        closeDatabaseConnection(c);
        List studentsOnAcceptedCourses = new LinkedList();
        Student stu = null;
        Course acceptedCourse = null;
        ListIterator acceptedCourseIterator = null;

        if(foundStudents == null) {
            return null;
        }

        ListIterator studentIterator = foundStudents.listIterator();
        while(studentIterator.hasNext()) {
            stu = (Student)studentIterator.next();
            acceptedCourseIterator = courses.listIterator();
            while(acceptedCourseIterator.hasNext()) {
                acceptedCourse = (Course)acceptedCourseIterator.next();
                if(stu.isOnCourse(acceptedCourse)) {
                    studentsOnAcceptedCourses.add(stu);
                    break;
                }
            }
        }

        return studentsOnAcceptedCourses;
    }
    
    /**
     * Retrieves a list of students, with data matching the specified search criteria, from the database.
     * If no matching students are found or if the name search string is shorter than 4 characters and student,
     * number search string is shorter than 3 characters,
     * null is returned. If both parameters contain data, the search is done based on the students name.
     * @param nameToSearch name (or part of name) of student to find
     * @param studentNumberToSearch student number (or part of it)
     */
    private List searchStudentsFromDatabase(String idToSearch, String nameToSearch, String studentNumberToSearch, Connection con) {
        List searchResults = new LinkedList();
        try {
            String query = "";
            PreparedStatement st;
            if (nameToSearch.length() >= 3) {
                query = "SELECT id FROM users WHERE LOWER(name) LIKE ?";
                st = con.prepareStatement(query);
                st.setString(1, "%" + nameToSearch.toLowerCase().trim() + "%");
            } else if (studentNumberToSearch.length() >= 2) {
                query = "SELECT id FROM users WHERE LOWER(studentNumber) LIKE ?";
                st = con.prepareStatement(query);
                st.setString(1, "%" + studentNumberToSearch.toLowerCase().trim() + "%");
            } else if (idToSearch.length() >= 2) {
                query = "SELECT id FROM users WHERE LOWER(id) LIKE ?";
                st = con.prepareStatement(query);
                st.setString(1, "%" + idToSearch.toLowerCase().trim() + "%");
            } else {
                return searchResults;
            }

            ResultSet results = st.executeQuery();

            while(results.next()) {
                Student student = getStudent(results.getString("id"));

                if (student != null) {
                    searchResults.add(student);
                }
            }

            return searchResults;

        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "getStudentsFromDatabase():", e);
        }

        return null;
    }


    /**
     * Populate a student object with results from all courses the student is registered to.
     *
     * @param s Student object
     * @param con
     */
    private void setResultsForStudent(Student s, Connection con) {
        try {
            if (preparedStatement == null || preparedStatement.getConnection() != con) {
                String query = "SELECT normalpoints, makeuppoints, eid, cid, rid, subs, rnumber, enumber, makeupfactor FROM "
                    + "((SELECT normalpoints, one.eid, one.cid, one.rid, subs FROM "
                    + "(SELECT COUNT(points) AS subs, eid, cid, rid FROM "
                    + "results where userid=? group by eid, cid, rid) AS one "
                    + "LEFT OUTER JOIN "
                    + "(SELECT MAX(points) AS normalpoints, eid, cid, rid FROM "
                    + "results where userid=? and (status="
                    + DataStorage.OK
                    + "or status="
                    + DataStorage.OK_MODIFIED
                    + ") group by eid, cid, rid) AS two "
                    + "USING(cid, rid, eid)) AS first "
                    + "LEFT OUTER JOIN "
                    + "(SELECT max(points) as makeuppoints, eid, results.cid, rid, makeupfactor FROM "
                    + "results,round WHERE round.id = rid AND userid=? and ( status="
                    + DataStorage.AFTER_DEADLINE
                    + " OR status="
                    + DataStorage.AFTER_DEADLINE_MODIFIED
                    + ") GROUP BY "
                    + "eid, results.cid, rid, makeupfactor) AS second USING(cid,rid,eid)) "
                    + "LEFT OUTER JOIN (SELECT cid, id as rid , eid, round.number as rnumber, exercises.number as enumber FROM round, exercises WHERE id=rid) AS nros "
                    + "USING(cid, rid, eid)";
                preparedStatement = con.prepareStatement(query);
            }

            preparedStatement.setString(1, s.getId());
            preparedStatement.setString(2, s.getId());
            preparedStatement.setString(3, s.getId());

            ResultSet rs = preparedStatement.executeQuery();
            while (rs.next()) {
                // int eid = rs.getInt(3);
                int cid = rs.getInt(4);
                Course c = getCourseFromCache(cid);

                if(c == null) {
                    c = getCourseFromDatabase(cid,con);
                    if (c == null) {
                        Logging.error(this, "Course "+cid+" not found");
                        continue;
                    }
                }

                // int rid = rs.getInt(5);
                int normalPoints = rs.getInt(1);
                int makeupPoints = rs.getInt(2);
                int subs = rs.getInt(6);
                float makeup = rs.getFloat(9);
                int rNro = rs.getInt(7);
                int eNro = rs.getInt(8);
                Results r = s.getResults(c);

                if (r == null) {
                    Logging.error(this, "Results on course " + c + " for student " + s + " is null, cannot assing points");
                    continue;
                }

                int points = normalPoints;
                if(normalPoints < (int)Math.ceil(makeupPoints*makeup)) {
                    points = (int)Math.ceil(makeupPoints*makeup);
                }

                r.setResults(rNro, eNro, points, subs);
            }
        } catch(Exception e) {
            e.printStackTrace();
            Logging.error(this, "Exception setting results for student "+s.getId(), e);
        }

    }

    /**
     * Gets a list containing all courses the given student is registered to.
     * 
     * @param studentId
     * @param priviledge
     * @param con
     * @return
     */
    private List getCoursesForStudent(String studentId, int priviledge, Connection con) {
        try {
            String query = "";
            PreparedStatement st;
            ResultSet rs = null;
            if(priviledge > Student.DEFAULT_ACCESS) {
                query = "SELECT * FROM Registrations WHERE Registrations.userId = ?" +
                " AND registrations.cId = " + 0;
                st = con.prepareStatement(query);
                st.setString(1, studentId);
                rs = st.executeQuery();
                if(rs.next()) {
                    query = "SELECT code, id FROM Course WHERE id > 0";
                    st = con.prepareStatement(query);
                } else {
                    query = "SELECT code, id FROM Course, Registrations WHERE Registrations.userId = ?" +
                     " AND Registrations.userpriviledge = " + priviledge + " AND Registrations.cId = Course.id";
                    st = con.prepareStatement(query);
                    st.setString(1, studentId);
                }
            } else {
                query = "SELECT code, id FROM Course, Registrations WHERE Registrations.userId = ?" +
                 " AND Registrations.userpriviledge = " + priviledge + " AND Registrations.cId = Course.id";
                st = con.prepareStatement(query);
                st.setString(1, studentId);
            }
            rs = st.executeQuery();
            List results = new LinkedList();
            while(rs.next()) {
                Course c = null;
                // String code = rs.getString(1);
                int id = rs.getInt(2);
                /*c = getCourseFromCache(id);
                if(c != null) {
                    results.add(c);
                    continue;
                }
                String name = rs.getString(2);
                c = getCourseContentsFromDatabase(code, name, id, con);*/
                // Just get the courses from the method that fetches
                // them from db in case they aren't in cache already
                c = getCourse(id);
                addCourse(c, id);
                results.add(c);
            }
            return results;
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting courses for student "+studentId, e);
        }
        return null;
    }


    /**
     * Gets the course content for the given course from database.
     * 
     * @param code
     * @param cid
     * @param con
     * @return
     */
    private Course getCourseContentsFromDatabase(String code, int cid, Connection con) {
	try {
	    String query = "SELECT name, lang, description, link FROM coursedescription WHERE cId = "+cid;
	    Statement st = con.createStatement();
	    ResultSet rs = st.executeQuery(query);
	    Course c = new Course(code);
	    c.setId(cid);
	    while(rs.next()) {
		c.addLocalizedData(rs.getString("lang"),
				   rs.getString("name"),
                    rs.getString("description"),
                    rs.getString("link"));
	    }
	    c.setId(cid);
	    query = "SELECT id FROM Round WHERE cid = "+cid +" order by number";
	    st = con.createStatement();
	    rs = st.executeQuery(query);
	    while(rs.next()) {
		int rid = rs.getInt(1);
		Round r = getRoundFromDatabase(rid, con);
		c.addRound(r);
	    }
	    List l = getAppletParametersFromDatabase(cid, 'c', con);
	    c.setAdditionalAppletParameters(l);
	    //Grading
	    query = "SELECT count(*) FROM coursegrading WHERE cId=" + cid;
	    st = con.createStatement();
	    rs = st.executeQuery(query);
	    int count = 0;
	    while(rs.next()) {
		count = rs.getInt("count");
	    }
	    query = "SELECT grade, minpoints FROM coursegrading WHERE cId=" + cid + " ORDER BY minpoints";
	    st = con.createStatement();
	    rs = st.executeQuery(query);
	    int[] minPoints = new int[count];
	    String[] grades = new String[count];
	    count = 0;
	    while(rs.next()) {
		minPoints[count] = rs.getInt("minpoints");
		grades[count] = rs.getString("grade");
		count++;
	    }
	    c.setGrades(minPoints, grades);
	    return c;
	} catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Exception getting course contents for course "+cid, e);
	}
	return null;
    }

    /**
     * Gets a round from the database
     * 
     * @param rid
     * @param con
     * @return
     */
    private Round getRoundFromDatabase(int rid,  Connection con) {
	try {
	    Round r = new Round();
	    String query = "";
	    Statement st;
	    ResultSet rs;
	    if(rid > 0) { 	    // normal round fetching
		query = "SELECT deadline, min, number, makeupfactor, open FROM round WHERE id=" + rid;
		st = con.createStatement();
		rs = st.executeQuery(query);
		while(rs.next()) {
		    r.setDeadLine(rs.getTimestamp("deadline"));
		    r.setMin(rs.getDouble("min"));
		    r.setRoundNumber(rs.getInt("number"));
		    r.setMakeupFactor(rs.getDouble("makeupfactor"));
		    r.setId(rid);
		    r.setOpen(rs.getTimestamp("open"));
		}	    
		query = "SELECT name, lang, description FROM rounddescription WHERE rid=" + rid;
		st = con.createStatement();
		rs = st.executeQuery(query);
		while(rs.next()) {
		    r.addLocalizedData(rs.getString("lang"),
				       rs.getString("name"),
				       rs.getString("description"));
		}


                /*
                 * select id, code, width, height, codebase, archive, hintpage,
                 * questiongroup_id, number, maxpoints, maxsubs from exercises,
                 * (select id, a.code, a.width, a.height, a.codebase, a.archive,
                 * a.hintpage, m.questiongroup_id from exercise FULL OUTER JOIN
                 * appletexercise AS a ON (a.eid = id) LEFT OUTER JOIN
                 * multiplechoiceexercise AS m ON (m.eid = id)) as ex where
                 * ex.id=exercises.eid and exercises.rid=101 ORDER BY number;
                 */

                /*query = "SELECT id, code, height, width, hintpage, codebase, archive, number,"+
		    " maxPoints, maxsubs FROM Exercise, Exercises WHERE Exercises.rId = " + rid + 
                " and exercises.eId = exercise.id ORDER BY number";*/

                query = "select id, type, code, width, height, codebase, archive, hintpage, " +
                        "number, maxpoints, maxsubs from exercises, " +
                        "(select id, type, a.code, a.width, a.height, a.codebase, a.archive, " +
                        "a.hintpage from exercise " +
                        "LEFT OUTER JOIN appletexercise AS a ON (a.eid = id)) as ex " +
                        "where ex.id=exercises.eid and exercises.rid=" + rid + " ORDER BY number;";
	    } else { // get all exercises to a single round not connected to a a course
                //query = "SELECT id,code,height,width,hintpage,codebase,archive FROM Exercise ORDER BY id";
                query = "select id, type, a.code, a.width, a.height, a.codebase, a.archive, " +
                    "a.hintpage from exercise " +
                    "LEFT OUTER JOIN appletexercise AS a ON (a.eid = id) ORDER BY id;";
	    }
	    st = con.createStatement();
	    rs = st.executeQuery(query);
	    //query = "SELECT name, description, lang FROM ExerciseDescription WHERE eid = ?";
	    //PreparedStatement ps = con.prepareStatement(query);
	    while(rs.next()) {
		int eid = rs.getInt(1);
		ExerciseTemplate et = getExerciseTemplate(eid);
		if(et == null) {
		    //ps.setInt(1, eid);
		    query = "SELECT name, description, lang FROM ExerciseDescription WHERE eid = "+eid + " ORDER BY lang";
		    st = con.createStatement();
		    ResultSet drs = st.executeQuery(query);
		    if(drs.next()) {
			String[] params = new String[8];
			params[0] = drs.getString(1);
			params[1] = drs.getString(2);
			params[2] = rs.getString("codebase");
			params[3] = rs.getString("archive");
			params[4] = rs.getString("code");
			params[5] = rs.getString("width");
			params[6] = rs.getString("height");
			params[7] = rs.getString("hintpage");
			query = "SELECT name, value FROM AppletParameter WHERE type = 'e' AND id = " +eid;
			st = con.createStatement();
			ResultSet prs = st.executeQuery(query);
			List l = new LinkedList();
			while(prs.next()) {
			    l.add(new AppletParameter(prs.getString(1), prs.getString(2)));
			}
            et = new ExerciseTemplate(params, l);
			et.setId(eid);
			l = getAppletParametersFromDatabase(eid, 'e', con);
			et.setAppletParameters(l);
			et.addLocalizedData(drs.getString("lang"), params[0], params[1]);
			while (drs.next())
			    et.addLocalizedData(drs.getString("lang"),
			            drs.getString("name"),
			            drs.getString("description"));
			addExerciseTemplate(et, eid);
		    } else {
		        System.err.println("Exercise description not found");
		        Logging.warn(this, "Exercise description not found for "+eid);
		    }
		}
		Exercise ex;
		if(rid > 0) {
		    ex = new Exercise(et, rs.getInt("maxPoints"), rs.getInt("number"), rs.getInt("maxSubs"));
		} else {
		    ex = new Exercise(et,0,0,0);
		}
		r.addExercise(ex);
	    }
	    List ls = getAppletParametersFromDatabase(rid, 'r', con);
	    r.setAdditionalAppletParameters(ls);
	    return r;
	} catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Exception getting round "+rid, e);
	}
	return null;
    }

    /**
     * Gets a course from the database.
     * 
     * @param courseId
     * @param con
     * @return
     */
    private Course getCourseFromDatabase(int courseId, Connection con) {
        try {
            String query = "SELECT code, id, minPoints, open, close, organization_id, organizer_uid FROM Course WHERE id = "+courseId+" ORDER BY id DESC";
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(query);

            if(rs.next()) {
                String code = rs.getString(1);
                int id = rs.getInt(2);
                double min = rs.getDouble("minPoints");
                Course c = getCourseContentsFromDatabase(code, id, con);
                c.setMin(min);
                c.setOpen(rs.getTimestamp("open"));
                c.setClose(rs.getTimestamp("close"));
                String organizerUId = rs.getString("organizer_uid");
                int oid = rs.getInt("organization_id");
                //if(organizerUId != null) XXX Causes recursion.. this has to be fixed in some other way!
                    //c.setOrganizer(getMonitor(organizerUId));

                if(oid > 0) {
                    c.setOrganization(getOrganizationFromDatabase(oid, con));
                }

                addCourse(c, id);
                return c;
            }
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting course ("+courseId+") from db", e);
        }

        return null;
    }

    /**
     * Gets a list containing all currently open courses.
     * 
     * @param con
     * @return List of courses
     */
    private List getOpenCourses(Connection con) {
	try {
	    SimpleDateFormat dform = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	    java.util.Date d = new java.util.Date();
	    String query = "SELECT code, id FROM Course WHERE open < '" + dform.format(d) + 
		"' and close > '" + dform.format(d) + "' ORDER BY id DESC";
	    Statement st = con.createStatement();
	    ResultSet rs = st.executeQuery(query);
	    List l = new LinkedList();
	    while(rs.next()) {
		int id = rs.getInt(2);
		Course c = getCourseFromCache(id);
		if(c == null) {
		    c = getCourseFromDatabase(id, con);
		}
		l.add(c);
	    }
	    return l;
	} catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Exception getting courses from db", e);
	}
	return null;
    }

    
    public Organization getOrganization(int oid) {
        Connection c = getDatabaseConnection();
        return getOrganizationFromDatabase(oid, c);
    }
    
    /**
     * Gets an organization-object from the database
     * @param oid The organization id
     * @param con The database connection to be used
     * @return Organization object
     */
    public Organization getOrganizationFromDatabase(int oid, Connection con) {
        Organization organization = new Organization(oid);
        String lang = null;
        try {
            String query = "SELECT id, shibid FROM organization WHERE id=" + oid;
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(query);
            if (rs.next()) {
                organization.setShibId(rs.getString("shibid"));
            }

            // Get organization description
            query = "SELECT orgid, lang, acronym, name, link, description FROM organizationdescription WHERE orgid=" + oid;
            st = con.createStatement();
            rs = st.executeQuery(query);
            while (rs.next()) {
                lang = rs.getString("lang");
                organization.addLocalizedAcronym(lang, rs.getString("acronym"));
                organization.addLocalizedName(lang, rs.getString("name"));
                organization.addLocalizedLink(lang, rs.getString("link"));
                organization.addLocalizedDescription(lang, rs.getString("description"));
            }
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting courses from db", e);
        }

        if(lang != null) {
            return organization;
        } else {
            return null;
        }
    }

    /**
     * Fetches the statistics for a certain course from the database
     * @param courseId Unique course id
     * @param con The database connection to be used
     */ 
    private Statistics getStatisticsFromDatabase(int courseId, Connection con) {
	try {
	    Statistics sta = new Statistics(courseId);
	    String query = "SELECT count(*) FROM registrations where cid="+courseId +" AND userpriviledge="+STUDENT;
	    Statement st = con.createStatement();
	    ResultSet rs = st.executeQuery(query);
	    if(rs.next()) {
		int r = rs.getInt(1);
		sta.setNroStudents(r);
	    } else {
		// XXX Error
		System.err.println("Could not retrieve the number of students registered to course " + courseId);
	    }
	    rs.close();
	    query = "SELECT count(*) FROM registrations where cid="+courseId +" AND userpriviledge="+UNREGISTERED;
	    st = con.createStatement();
	    rs = st.executeQuery(query);
	    if(rs.next()) {
		int r = rs.getInt(1);
		sta.setNroUnregistered(r);
	    } else {
		// XXX Error
		System.err.println("Could not retrieve the number of unregistered students for course " + courseId);
	    }
	    st = con.createStatement();
	    // first create the statistics
	    query = " SELECT round.id, round.number, exercises.eid, exercises.number FROM round,exercises " + 
		"WHERE cid="+courseId + " AND round.id = exercises.rid ORDER BY round.number, exercises.number";
	    rs = st.executeQuery(query);
	    int rnro = -1;
	    int enro = -1;
	    RoundStatistics rstat = null;
	    ExerciseStatistics estat = null;
	    while(rs.next()) {
		int tmp = rs.getInt(2);
		if(tmp > rnro) {
		    rnro = tmp;
		    enro = -1;
		    rstat = new RoundStatistics();
		    sta.addRound(rstat);
		    rstat.setRoundNumber(rnro);
		}
		tmp = rs.getInt(4);
		if(tmp > enro) {
		    enro = tmp;
		    estat = new ExerciseStatistics();
		    rstat.addExercise(estat);
		    estat.setExerciseNumber(enro);
		}
		//estat.addStatistics(rs.getString(5), rs.getInt(6));
	    }
	    rs.close();
	    // then add the statistics themselves;
	    query = "SELECT round.id, round.number, exercises.eid, exercises.number, results.userid, results.points " + 
		"FROM round, exercises, results " +
		"WHERE round.cid = " + courseId + " AND results.cid = " + courseId + " AND round.id = exercises.rid AND " +
		"exercises.eid = results.eid AND round.id = results.rid " +
        "AND NOT (results.status =" + DataStorage.INVALID_FINGERPRINT + " OR results.status =" + DataStorage.INVALID_FINGERPRINT_MODIFIED +") " +
		"ORDER BY round.number, exercises.number, results.userid";
	    rs = st.executeQuery(query);
	    rnro = enro = -1;
	    rstat = null;
	    estat = null;
	    while(rs.next()) {
		int tmp = rs.getInt(2);
		if(tmp > rnro) {
		    rnro = tmp;
		    enro = -1;
		    rstat = sta.getRoundStatistics(rnro-1);
		}
		tmp = rs.getInt(4);
		if(tmp > enro) {
		    enro = tmp;
		    estat = rstat.getExercise(enro-1);
		}
		estat.addStatistics(rs.getString(5), rs.getInt(6));
	    }

	    return sta;
	}catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Exception getting course statistiscs for course "+courseId+" from db", e);
	}
	return null;
    }

    /**
     * Fetches a ResultSet containing the results returned between d1 and d2 for a certain course.
     * The set contains all results sent after d1 and before d2 at the course with the given id.
     */
    private ResultSet getResultsByDateFromDatabase(java.util.Date d1, java.util.Date d2, int id, Connection con) {
	SimpleDateFormat dform = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	String query = "SELECT * FROM results where tstamp BETWEEN '" + dform.format(d1) + "' and '" + dform.format(d2) + 
	    "' and cid = " + id + " ORDER BY tstamp";
	try {
	    ResultSet rs = executeQuery(query, con);
	    return rs;
	} catch(SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Exception fetching results by date");
	}
	return null;
    }

    /** 
     * Opens a database connection. The connection must be closed
     * using closeDatabaseConnection. 
     */
    private static Connection getDatabaseConnection() {
	try { 
	    String driver = System.getProperty("jdbc.driver");
	    // Needs to be done on the first time only??
	    Class.forName(driver);


	    String dbUrl = System.getProperty("db.url");
	    String username = System.getProperty("db.username");
	    String passwd = System.getProperty("db.password");
	    return DriverManager.getConnection(dbUrl, username, passwd); 
	} catch (ClassNotFoundException e) {
	    e.printStackTrace();
	    Logging.fatal(getInstance(), "Database driver not found", e);
	    return null;
	} catch (SQLException e) {
	    e.printStackTrace();
	    Logging.fatal(getInstance(), "Database connection couldn't be opened", e);
	    return null;
	}
    }

    /** 
     * Returns the ResultSet of the given query or null if an 
     * exception occurs.
     */
    private ResultSet executeQuery(String query) throws SQLException {
	Connection con = getDatabaseConnection();
	ResultSet result = executeQuery(query,con);
	con.close();
	return result;
    }

    /** 
     * Returns the ResultSet of the given query or null if an 
     * exception occurs.
     */
    private ResultSet executeQuery(String query, Connection con) throws SQLException {
	Statement stmt = con.createStatement();
	ResultSet result = stmt.executeQuery(query);
	return result;
    }
    
    /**
     * Inserts the given result into the database.
     * 
     * @param studentId
     * @param courseId
     * @param round
     * @param exercise
     * @param points
     * @param fileName
     * @param status
     * @return
     */
    private boolean addResult(String studentId,
                    String courseId, int round,
                    int exercise, int points, String fileName, int status)
    {
        try {
            Connection con = getDatabaseConnection();
            Statement stmt = con.createStatement();
            ResultSet ids = stmt.executeQuery("SELECT round.id, eid " +
                            "FROM round, exercises " +
                            "WHERE cid="+courseId+" AND " +
                            "round.id=exercises.rid AND round.number="+round+" AND exercises.number="+exercise+";");
            int rId = -1;
            int eId = -1;
            if (ids.next()) {
                rId = ids.getInt("id");
                eId = ids.getInt("eId");
            } else {
                System.err.println("Round or exercise number not found");
                Logging.error(this, "Round ("+round+") or exercise number ("+exercise+") not found");
            }
            java.util.Date currTime = Calendar.getInstance().getTime();
            DateFormat df = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            String updateQuery = "INSERT INTO RESULTS VALUES(" + points + ",'" + df.format(currTime) + "',?," + eId + ",?," + rId + ", " + status + ", ?);";
            PreparedStatement pst = con.prepareStatement(updateQuery);
            pst.setString(1, studentId);
            pst.setInt(2, Integer.parseInt(courseId));
            pst.setString(3, fileName);
            int result = pst.executeUpdate();
            con.close();
            return true;
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Couldn't add result ("+studentId+" - "+courseId+"-"+round+"-"+exercise+" points: "+points+")", e);
            return false;
        }
    }

    public boolean updateResult(String studentId, Course course, Round round,
            Exercise exercise, int points, java.util.Date timestamp, int status)
    {
        if(status < 10) { //Changes made from monitor ui should be traceable by the status flag (normal status + 10)
            return false;
        }

        if(points > exercise.getMax()) {
            return false;
        }

        boolean success = editResult(studentId, course.getId(), round.getId(), exercise.getExerciseTemplate().getId(), points, timestamp, status);
        if (!success) {
            return false;
        } else {
            //Student st = getStudent(studentId);
            //st.editResult(course,round,exercise,points);
            return true;
        }
    }


    private synchronized boolean editResult(String studentId,
              int courseId, int roundId,
              int exerciseId, int points, java.util.Date timestamp, int status) {
        try {
            Connection con = getDatabaseConnection();

            DateFormat df = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            String editQuery = "UPDATE results SET points=" + points + ", status=" + status + 
                " WHERE userid=? AND cid=" + courseId + " AND rid=" + roundId + " AND eid=" + exerciseId + " AND tstamp='" + df.format(timestamp) + "';";

            PreparedStatement stmt = con.prepareStatement(editQuery);
            stmt.setString(1, studentId);
            int result = stmt.executeUpdate();
            con.close();
            return true;
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Couldn't edit result ("+studentId+" - "+courseId+"-"+roundId+"-"+exerciseId+" points: "+points+" status: " + status+ ")", e);
            return false;
        }
    }

    private List getAppletParametersFromDatabase(int id, char type, Connection con) {
        List l = new LinkedList();
        try {
            String query = "SELECT name, value FROM AppletParameter WHERE type='"+type+"' AND ( id="
            + id + " OR id=0)";
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(query);
            while(rs.next()) {
                String name = rs.getString(1);
                String value = rs.getString(2);
                l.add(new AppletParameter(name, value));
            }
            return l;
        } catch(SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting appletparameters from db", e);
        }
        return null;
    }

    /**
     * Get a list of available locales. The list is populated by selecting all distinct locales
     * from ExerciseDescription table.
     * @param con
     * @return List of locales
     */
    private List getLocales(Connection con) {
        List locales = new LinkedList();
        try {
            String query = "SELECT DISTINCT lang FROM ExerciseDescription ORDER BY lang;";
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(query);
            while(rs.next()) {
                String language = rs.getString("lang");
                locales.add(new Locale(language));
            }

            return locales;
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception getting locales from db", e);
        }
        return null;
    }

    /**
     * Returns true, if registrations with the given student ID 
     * exist. Otherwise, returns false.
     */

    public boolean hasRegistrations(String studentId) {
        try {
            Connection con = getDatabaseConnection();
            PreparedStatement stmt = con.prepareStatement("SELECT cid FROM registrations " +
                    "WHERE userid=?");
            stmt.setString(1, studentId);
            ResultSet rs = stmt.executeQuery();
            while (rs.next()) {
                int cId = rs.getInt("cid");
                if (cId >= 0) {
                    return true;
                }
            }
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Couldn't get registrations for student "+studentId, e);
            return false;
        }
        return false;
    }

    public String getParameterValue(String name, String type, int id) {
        try {
            Connection con = getDatabaseConnection();
            PreparedStatement stmt = con.prepareStatement("SELECT value FROM parameter " +
                    "WHERE name=? and type=? and id="+id);
            stmt.setString(1, name);
            stmt.setString(2, type);
            ResultSet rs = stmt.executeQuery();
            if (rs.next())
            return rs.getString("value");
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Couldn't get parameter " + name + "(" + type + ", " + id + ")");
            return "";
        }
        return "";
    }

    /**
     * Get the link target associated with the given link name
     * @param linkName
     * @return
     */
    public String getLinkTarget(String linkName) {
	try {
	    Connection con = getDatabaseConnection();
	    PreparedStatement stmt = con.prepareStatement("SELECT target FROM link " +
                "WHERE name=?;");
        stmt.setString(1, linkName);
	    ResultSet rs = stmt.executeQuery();
	    if (rs.next())
		return rs.getString("target");
	    else 
		return "";
	} catch (SQLException e) {
	    e.printStackTrace();
	    Logging.error(this, "Couldn't get link " + linkName);
	    return "";
	}
    }

    /**
     * Gets course-independent messages (for the TRAKLA2 system) in the given language. Eg. information
     * about service-breaks and such.
     * 
     * @param language used language
     * @return message
     */
    public String getMessage(String language) {
        try {
            Connection con = getDatabaseConnection();
            PreparedStatement stmt = con.prepareStatement("SELECT value FROM parameter " +
                    "WHERE name=? AND type='c' AND id=0;");
            stmt.setString(1, "message_" + language);
            ResultSet rs = stmt.executeQuery();
            if (rs.next()) {
                return rs.getString("value");
            } else {
                return "";
            }
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Couldn't get message!");
            return "";
        }
    }

    /**
     * Sets the default course for a student
     * @param studentNumber the user id of the student
     * @param cid of course
     */
    public void setStudentDefaultCourse(String studentId, int cid) {
        try {
            Connection con = getDatabaseConnection();
            // make sure student is in db
            Student s = getStudentFromDatabase(studentId, con);
            String update = "UPDATE users SET defcourse ='" + cid + "' WHERE id=?;";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, studentId);
            st.executeUpdate();
            closeDatabaseConnection(con);
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception setting student latest course ("+studentId+") =" + cid, e);
        }
    }

    /**
     * Update the exercise applets width and height for the given exercise template id
     * @param id of the exercise template to update
     * @param appletWidth the new width
     * @param appletHeight the new height
     * @return true if succefully changed, false otherwise
     */
    public boolean setExerciseAppletSize(int id, int appletWidth, int appletHeight) {
        try {
            Connection con = getDatabaseConnection();
            Statement st = con.createStatement();
            String update = "UPDATE appletexercise SET width =" + appletWidth + ", height =" + appletHeight +" WHERE eid='" + id + "';";
            st.executeUpdate(update);
            closeDatabaseConnection(con);
            return true;
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Exception setting exercise applet size. Id= "+id+ ", new width=" + appletWidth + " new height: " + appletHeight + ".", e);
            return false;
        }
    }
    
    /**
     * Returns the user id associated with the given hash string. 
     * 
     * @param hash hash string
     * @return user id or null if not found
     */
    public String getRecoveryPageData(String hash) {
        String userId = null;
        try {
            Connection con = getDatabaseConnection();
            String query = "SELECT id FROM recoverpassword WHERE hash=?;";
            PreparedStatement st = con.prepareStatement(query);
            st.setString(1, hash);
            ResultSet rs = st.executeQuery();
            if(rs.next())
                userId = rs.getString(1);
            closeDatabaseConnection(con);
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Error getting user id associated with hash value " + hash + ".", e);
        }
        return userId;
    }
    
    /**
     * Adds a recovery page hash for the given user id.
     * 
     * @param hash hash data string
     * @param userId user id
     */
    public synchronized void setRecoveryPageData(String hash, String userId) {
        if(hash == null || userId == null)
            return;
        try {
            Connection con = getDatabaseConnection();
            String update = "INSERT INTO recoverpassword VALUES (?, ?);";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, userId);
            st.setString(2, hash);
            st.executeUpdate();
            closeDatabaseConnection(con);
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Error adding values to recoverpassword table: " + userId + ", " + hash + ".", e);
        }
    }

    /**
     * Removes recovery page data for the given user id.
     * @param userId user id
     */
    public synchronized void removeRecoveryPageData(String userId) {
        try {
            Connection con = getDatabaseConnection();
            String update = "DELETE FROM recoverpassword WHERE id=?;";
            PreparedStatement st = con.prepareStatement(update);
            st.setString(1, userId);
            st.executeUpdate();
            closeDatabaseConnection(con);
        } catch (SQLException e) {
            e.printStackTrace();
            Logging.error(this, "Error removing data from recoverpassword table, user id: " + userId + ".", e);
        }
    }
}
