// Copyright 2007 Tapio Auvinen, Erik Fallenius, Ville Karavirta,
// Jussi Nikander, and Harri Salonen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * 
 */
package cexer;

import java.util.GregorianCalendar;
import java.util.List;
import java.util.LinkedList;
import org.apache.tapestry.BaseSessionStoreOptimized;
import cexer.util.Logging;

/**
 * This is the session-object for the monitor interface. It is used to store all
 * information concerning a certain visit to the TRAKLA2 web application. <br>
 * StudentSession.java and MonitorSession.java replace Visit.java in the new
 * TRAKLA2-revision (summer 2006).<br>
 * 
 * @author cefallen
 * 
 */
public class MonitorSession extends BaseSessionStoreOptimized {

    /** 
     * Unique string identifying the student. This was called username in older trakla2 version. 
     * Identifiers "id" and "userid" are used in the database.
     */
    private String userId;

    private String name;

    private String password;

    private Student student;

    private Student monitor;

    private Exercise exercise;

    private Course course;

    private Round round;

    private ExerciseTemplate exerciseTemplate;

    // fields relevant to monitors only
    private GregorianCalendar start;

    private GregorianCalendar end;

    private int activityViewMode;

    private boolean[] roundIndicator;

    private boolean isLoggedIn;

    private Statistics courseStatistics;

    private List studentResults;

    // view modes
    public static final int NONE = -1;

    public static final int DAY = 0;

    public static final int WEEK = 1;

    public static final int COURSE = 2;

    public static final int DAY_AVERAGE = 3;

    public MonitorSession() {
        isLoggedIn = false;
        monitor = null;
        //studentsFound = new LinkedList();
        studentResults = new LinkedList();
    }
    
    public boolean getIsLoggedIn() {
        return isLoggedIn;
    }
    
    public void setIsLoggedIn(boolean state) {
        isLoggedIn = state;
        markSessionStoreNeeded();        
    }

    public boolean[] getRoundIndicator() {
        if (roundIndicator == null) {
            roundIndicator = new boolean[getCourse().getRoundList().size() + 1];
        } else if (roundIndicator.length != getCourse().getRoundList().size()) {
            boolean[] oldRounds = roundIndicator;
            roundIndicator = new boolean[getCourse().getRoundList().size() + 1];
            for (int i = 0; i < roundIndicator.length - 1
                    && i < oldRounds.length - 1; i++) {
                roundIndicator[i] = oldRounds[i];
            }
            roundIndicator[roundIndicator.length - 1] = oldRounds[oldRounds.length - 1];
        }
        markSessionStoreNeeded();
        return roundIndicator;
    }

    public String getUserId() {
        return userId;
    }

    public void setUserId(String value) {
        userId = value;
        markSessionStoreNeeded();
    }

    public String getName() {
        return name;
    }

    public void setName(String value) {
        name = value;
        markSessionStoreNeeded();
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String value) {
        password = value;
        markSessionStoreNeeded();
    }

    public Student getStudent() {
        if (monitor == null) {
            if (userId == null) {
                return null;
            }
            return DataStorage.getInstance().getStudent(userId);
        } else {
            return monitor;
        }
    }

    public Student getMonitor() {
        return monitor;
    }

    public void setMonitor(Student s) {
        if (s.isInstructor()) {
            monitor = s;
            markSessionStoreNeeded();
        }
    }

 
    public boolean isInstructor() {
        if (monitor != null && monitor.isInstructor())
            return true;
        return false;
    }
  
    
    public boolean isAdmin() {
        if (monitor != null && monitor.isAdmin())
            return true;
        return false;
    }

    public Student getMonitoredStudent() {
        return student;
    }

    public void setMonitoredStudent(Student value) {
        student = value;
        markSessionStoreNeeded();
    }

    public Exercise getExercise() {
        return exercise;
    }

    public void setExercise(Exercise ex) {
        exercise = ex;
        markSessionStoreNeeded();
    }

    public Course getCourse() {
        if (course == null) {
            Student stu = getStudent();
            if (stu == null)
                return null;
            if (stu.getDefaultCourse() != null) {
                setCourse(stu.getDefaultCourse());
            } else {
                setCourse((Course) stu.getCourseList().get(0));
            }
        }
        return course;
    }

    public void setCourse(Course c) {
        if (course != null && c != null && course.getId() != c.getId()) {
            course = c;
            studentResults = null;
            courseStatistics = null;
        }
        if (course == null && c != null) {
            course = c;
            studentResults = null;
            courseStatistics = null;
        }
        markSessionStoreNeeded();
    }

    public void clearCourse() {
        course = null;
        markSessionStoreNeeded();
    }

    public Round getRound() {
        return round;
    }

    public void setRound(Round r) {
        round = r;
        markSessionStoreNeeded();
    }

    public void setExerciseTemplate(ExerciseTemplate et) {
        exerciseTemplate = et;
        markSessionStoreNeeded();
    }

    public ExerciseTemplate getExerciseTemplate() {
        return exerciseTemplate;
    }

    public GregorianCalendar getStart() {
        return start;
    }

    public void setStart(GregorianCalendar d) {
        start = d;
        markSessionStoreNeeded();
    }

    public GregorianCalendar getEnd() {
        return end;
    }

    public void setEnd(GregorianCalendar d) {
        end = d;
        markSessionStoreNeeded();
    }

    public int getActivityViewMode() {
        return activityViewMode;
    }

    public void setActivityViewMode(int i) {
        activityViewMode = i;
        markSessionStoreNeeded();
    }

    public void clearFields() {
        userId = null;
        name = null;
        password = null;
        student = null;
        monitor = null;
        exercise = null;
        course = null;
        round = null;
        exerciseTemplate = null;
        start = null;
        end = null;
        activityViewMode = NONE;
        roundIndicator = null;
        markSessionStoreNeeded();
    }

    /**
     * Get the students statistics for current course
     * @return Statistics
     */
    public Statistics getStatistics() {
        if (courseStatistics == null) {
            DataStorage ds = DataStorage.getInstance();
            courseStatistics = ds.getStatisticsForCourse(getCourse().getId());
        }
        markSessionStoreNeeded();
        return courseStatistics;
    }

    /**
     * Get student results for current course
     * @return List of results
     */
    public List getStudentResults() {
        if (studentResults == null) {
            DataStorage ds = DataStorage.getInstance();
            Logging.info(this, "monitor session calling getCourseResults()");
            studentResults = ds.getCourseResults(getCourse());
            Logging.info(this, "monitor session getCourseResults()");
        }
        markSessionStoreNeeded();
        return studentResults;
    }
}
