// Copyright 2007 Tapio Auvinen, Erik Fallenius, Ville Karavirta,
// Jussi Nikander, and Harri Salonen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cexer;

import java.util.List;
import java.util.LinkedList;

public class Results {

    private List results;
    private Course course;
    private List submissions;
    
    public Results(Course c) {
	course = c;
	results = createResults(c);
	submissions = createSubmissions(c);
    }

    private LinkedList createResults(Course c) {
	List rounds = c.getRoundList();
	LinkedList rl = new LinkedList();
	for(int i = 0; i < rounds.size(); i++) {
	    Round r = (Round)rounds.get(i);
	    List ex = r.getExerciseList();
	    int[] rs;
	    rs = new int[ex.size()];
	    rl.add(rs);
	}
	return rl;
    }
    
    private LinkedList createSubmissions(Course c) {
	List rounds = c.getRoundList();
	LinkedList rl = new LinkedList();
	for(int i = 0; i < rounds.size(); i++) {
	    Round r = (Round)rounds.get(i);
	    List ex = r.getExerciseList();
	    int[] rs;
	    rs = new int[ex.size()];
	    rl.add(rs);
	}
	return rl;
    }

    public int[] getRound(int round) {
	if(round > 0 && round <= course.getRoundList().size()) {
	    if(round-1 >= results.size()) {
		while(round-1 >= results.size()) 
		    results.add(null);
	    }
	    int[] rr = (int[])results.get(round-1);
	    if(rr == null) {
		Round r = (Round) course.getRoundList().get(round-1);
		int rSize = r.getExerciseList().size();
		if(rSize > 0) {
		    rr = new int[rSize];
		    for(int i = 0; i < rr.length; i++) rr[i] = 0;
		    results.set(round-1,rr);
		} else {
		    return null;
		}
	    }
	    return rr;
	} else {
	    return null;
	}
    }


    /**
     * First round of a course is assumed to be round number 1.
     * First exercise of a round is assumed to be round number 1.
     * Points must be a positive value.
     */
    public void setResult(int round, int exercise, int points) {
	int[] rr = getRound(round);
	if(rr == null)
	    return;
	addSubmission(round, exercise);
	if(exercise > 0 && points > 0) {
	    if(exercise > rr.length) {
		rr = setLength(exercise, rr);
		results.set(round-1,rr);
	    }
	    if(rr[exercise-1] < points)
		rr[exercise-1] = points;
	}
    }

    /** 
     * Sets the result for given round and exercise. The new value of submissions
     * for the exercise is subCount and the value of points is maxPoints.
     */
    public void setResults(int round, int exercise, int maxPoints, int subCount) {
	int[] rr = getRound(round);
	int[] subs = getSubmissions(round);
	if (rr == null || subs == null)
	    return;
	if (exercise > 0 && subCount > 0) {
	    if (exercise > rr.length) {
		rr = setLength(exercise, rr);
		results.set(round-1, rr);
	    }
	    if (exercise > subs.length) {
		subs = setLength(exercise, subs);
		submissions.set(round-1, subs);
	    }
	    subs[exercise-1] = subCount;
	    rr[exercise-1] = maxPoints;
	}
    }

    private void addSubmission(int round, int exercise) {
	int[] subs = getSubmissions(round);
	if (subs == null)
	    return;
	if (exercise > 0) {
	    if(exercise > subs.length) {
		subs = setLength(exercise, subs);
		submissions.set(round-1,subs);
	    }
	    subs[exercise-1]++;
	}
    }

    private int[] setLength(int newLength, int[] rr) {
	if(newLength > rr.length) {
	    int oldLength = rr.length;
	    int[] nr = new int[newLength];
	    int i = 0;
	    for(; i < rr.length; i++) nr[i] = rr[i];
	    for(; i < nr.length; i++) nr[i] = 0;
	    return nr;
	}
	return rr;
    }

    /**
     * See setResult
     */
    public int getResult(int round, int exercise) {
	int[] rr = getRound(round);
	if(rr == null)
	    return 0;
	if(exercise > 0) {
	    if(exercise > rr.length) {
		rr = setLength(exercise,rr);
		results.set(round-1,rr);
	    }
	    //System.out.println("Round " +round+ " exercise " + exercise + " : " + rr[exercise-1]);
	    return rr[exercise-1];
	}
	return 0;
    }

    private int[] getSubmissions(int round) {
	if(round > 0) {
	    if(round-1 > submissions.size()) {
		while(round-1 > submissions.size()) 
		    submissions.add(null);
	    }
	    int[] rr = (int[])submissions.get(round-1);
	    if(rr == null) {
		Round r = (Round)course.getRoundList().get(round-1);
		int rSize = r.getExerciseList().size();
		if(rSize > 0) {
		    rr = new int[rSize];
		    for(int i = 0; i < rr.length; i++) rr[i] = 0;
		    submissions.set(round-1,rr);
		} else {
		    return null;
		}
	    }
	    return rr;
	} else {
	    return null;
	}
    }
    
    public int getTotalSubmissions(int round) {
        int[] submissions = getSubmissions(round);
        int totalSubmissions = 0;
        if(submissions != null) {
            for(int i = 0; i < submissions.length; i++)
                totalSubmissions = totalSubmissions + submissions[i];
            return totalSubmissions;
        } else return 0;
    }

    public int getSubmissions(int round, int exercise) {
	int[] rr = getSubmissions(round);
	if(rr == null)
	    return 0;
	if(exercise > 0) {
	    if(exercise > rr.length) {
		rr = setLength(exercise,rr);
		submissions.set(round-1,rr);
	    }
	    return rr[exercise-1];
	}
	return 0;
    }

    public Course getCourse() {
	return course;
    }

    public int getRoundResult(int round) {
        if(round < 0 || round > course.getRoundList().size())
            return 0;
	int[] rnd = getRound(round);
	int res = 0;
	for(int i = 0; i< rnd.length; i++) {
	    res += rnd[i];
	}
	return res;
    }

//      public int[] getRoundResults(int round) {
//  	int[] rr = getRound(round-1);
//  	if(rr == null) {
//  	    return new int[0];
//  	}
//  	return rr.toArray();
//      }

}
