// Copyright 2007 Tapio Auvinen, Erik Fallenius, Ville Karavirta,
// Jussi Nikander, and Harri Salonen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cexer.pages;

import java.util.Locale;
import javax.servlet.http.HttpServletRequest;
import org.apache.tapestry.RedirectException;
import org.apache.tapestry.PageRedirectException;
import org.apache.tapestry.IRequestCycle;
import org.apache.tapestry.event.PageEvent;
import org.apache.tapestry.event.PageBeginRenderListener;
import org.apache.tapestry.services.RequestGlobals;
import org.apache.tapestry.annotations.InjectObject;


import cexer.*;
import cexer.util.Checker;
import cexer.util.Logging;


/** Provides dynamic behaviour for the LoginShib page */
public abstract class LoginShibPage extends SecurePage implements PageBeginRenderListener {

    /*
     * Needed to get HTTP headers
     */
    public abstract RequestGlobals getRequestGlobals();

    public abstract boolean isWarning();

    public abstract void setWarning(boolean b);

    public abstract String getWarningText();

    public abstract void setWarningText(String s);

    public abstract boolean isLoginSuccessful();
    
    public abstract void setLoginSuccessful(boolean succesful);
    
    // Set to false if, for example, student number is missing
    public abstract void setInformationComplete(boolean complete);
    
    public abstract boolean isInformationComplete();
    
    /**
     * Fetches the injected student-session object.
     */
    public abstract StudentSession getStudentSession();

    
    
    /**
     * Checks the shibboleth headers when the page is shown. If the user is
     * not authenticated, he is redirected to the login page of his home
     * university.
     */
    public void pageBeginRender(PageEvent event) {
        setLoginSuccessful(false);

        String principalName = getRequestGlobals().getRequest().getHeader("Shib_EP_PrincipalName");
        //Logging.info(this, "Shib_EP_PrincipalName: " + principalName);
        
        if (principalName == null) {
            // If shibboleth is disabled, redirect to traditional login page
            Logging.warn(this, "Shibboleth is disabled but somebody is trying to use LoginShib page");
            throw new PageRedirectException("Login");
        } else if (principalName.length() < 1) {
            // If the user is not authenticated, redirect to WAYF
            String shibUrl = determineShibUrl();
            //Logging.info(this, "User is not authenticated. Redirect to " + shibUrl);

            throw new RedirectException(shibUrl);
        }

        //Logging.info(this, "User is authenticated as " + principalName);

        // There is a principalName, so the user is authenticated.
        // Load user and course.
        DataStorage ds = DataStorage.getInstance();
        StudentSession session = getStudentSession();
        Student student = ds.getStudent(principalName);

        String cId = getRequestGlobals().getRequest().getParameter("cid");
        Course course = null;
        if (cId != null && cId.length() > 0) {
            try {
                course = ds.getCourse(Integer.parseInt(cId));
                //if (course == null) {
                    //Logging.warn(this, "Invalid course id as a parameter for RegisterShib page: " + cId);
                //}
            } catch (NumberFormatException e) {
                Logging.warn(this, "Invalid course id as a parameter for LoginShib page: " + cId);
            }
        }

        
        // Decide where to direct the user
        if (course != null) {
            // We have cId
            if (student != null && student.isOnCourse(course)) {
                // User is registered, so let him in
                //Logging.info(this, "User is registered. Let him in and set course to " + course);
                updateShibInfo(student);

                session.setCourse(course);
                ds.setStudentDefaultCourse(principalName, course.getId());

                loginUser(session, student);
            } else {
                // User is not registered, so register him.
                //Logging.info(this, "User is not registered. Registering.");
                if (student == null) {
                    student = createUser(principalName);
                    
                    if (student == null) {
                        return;
                    }
                }
                
                if (registerUser(student, course)) {
                    loginUser(session, student);
                }
            }
        } else {
            // We don't have cId
            if (student != null && student.getAllCourses().size() > 0) {
                // User is registered to the system, so let him in
                //Logging.info(this, "User is registered. Let him in and use the default course.");
                updateShibInfo(student);
                loginUser(session, student);
            } else {
                // User is not registered and no course is specified. Show an error page.
                //Logging.info(this, "User is not registered and no course is specified.");
                setWarningText(getMessage("error-not-registered"));
            }
        }
    }

    private void loginUser(StudentSession session, Student student) {
        if (session == null) {
            Logging.error(this, "LoginShibPage::loginUser: Session null.");
            setWarningText(getMessage("error-login-failed"));
            return;
        }
        if (student == null) {
            Logging.info(this, "LoginShibPage::loginUSer: Student null.");
            setWarningText(getMessage("error-login-failed"));
            return;
        }
        
        Logging.info(this, "Logging in student " + student);
        
        
        session.clearFields();
        
        // Get logout url 
        String logoutUrl = getRequestGlobals().getRequest().getHeader("Shib_logout-url");
        if (logoutUrl != null && logoutUrl.length() > 0) {
            session.setLogoutUrl(logoutUrl);
        }
        
        session.setIsLoggedIn(true);
        session.setAuthMethod(StudentSession.SHIBBOLETH);
        session.setUserId(student.getId());
        session.setLogoutUrl(logoutUrl);
        getEngine().setLocale(new Locale(student.getLanguage()));
        cexer.util.Logging.info(this, "Student " + student.getId() + " logged in.");
        
        // If information is not complete, redirect to Settings page
        if (student.getStudentNumber() == null || student.getStudentNumber().length() < 1) {
            setInformationComplete(false);
        } else {
            setInformationComplete(true);
        }
        
        setLoginSuccessful(true);
    }

    /**
     * 
     *
     */
    private Student createUser(String principalName) {
        //Create new student
        Logging.info(this, "Creating student " + principalName);
        Student student = new Student(principalName);
        DataStorage ds = DataStorage.getInstance();

        // Get shibInfo
        StudentShibInfo shibInfo = readShibInfo();
        student.setName(shibInfo.getCommonName());
        student.setEmail(shibInfo.getEmail());
        student.setStudentNumber(shibInfo.getStudentNumber());
        student.setLanguage(shibInfo.getLang());
        
        // Add new student to database
        if (!ds.addStudent(student)) {
            setWarningText(getMessage("error-adding-student"));
            Logging.error(this, "Error when adding student " + principalName);
            return null;
        }
        
        // Add shibinfo to database
        ds.saveStudentShibInfo(shibInfo);
        
        return student;
    }
    
    /**
     * Registers a student to a course.
     * @param student
     * @param course
     * @return
     */
    private boolean registerUser(Student student, Course course) {
        if (student == null) {
            Logging.error(this, "LoginShibPage::registerUser: Student null");
            return false;
        }
        if (course == null) {
            Logging.error(this, "LoginShibPage::registerUser: Course null");
            return false;
        }
        if (!course.isOpen()) {
            Logging.error(this, "Cannot register " + student + ". Course is not open.");
            setWarningText(getMessage("error-course-closed"));
            return false;
        }
        
        DataStorage ds = DataStorage.getInstance();

        student.addCourse(course);
        student.setDefaultCourse(course);
        ds.registerStudentToCourse(student.getId(), course.getId());
        ds.setStudentDefaultCourse(student.getId(), course.getId());
        getEngine().setLocale(new Locale(student.getLanguage()));
        
        cexer.util.Logging.info("Registered " + student.getId() + " to " + course);

        return true;
    }

    private StudentShibInfo readShibInfo() {
        // Read shibboleth headers
        HttpServletRequest request = getRequestGlobals().getRequest();
        
        String principalName = request.getHeader("Shib_EP_PrincipalName");

        String gender = request.getHeader("Shib_schac_Gender");
        String organization = request.getHeader("Shib_schac_HomeOrganization");
        String organizationName = request.getHeader("Shib_o");
        String studentNumber = parseShibStudentNumber(request.getHeader("Shib_schac_PersonalUniqueCode"));

        //cexer.util.Logging.info("Shib_schac_PersonalUniqueCode: " + request.getHeader("Shib_schac_PersonalUniqueCode"));

        String name = request.getHeader("Shib_CN");
        String nickName = request.getHeader("Shib_DisplayName");
        String givenName = request.getHeader("Shib_GIVENNAME");
        String surname = request.getHeader("Shib_sn");
        String email = request.getHeader("Shib_mail");
        String lang = request.getHeader("Shib_preferredLanguage");
        
        // Get new info from Shibboleth
        StudentShibInfo info = new StudentShibInfo(principalName);
        if (gender.equals("1")) {
            info.setGender("m");
        } else if (gender.equals("2")) {
            info.setGender("f");
        } else {
            info.setGender("u");
        }
        info.setOrganization(organization);
        info.setOrganizationName(organizationName);
        info.setStudentNumber(studentNumber);
        info.setCommonName(name);
        info.setDisplayName(nickName);
        info.setGivenName(givenName);
        info.setSurname(surname);
        info.setEmail(email);
        info.setLang(lang);
        
        return info;
    }
    
    private void updateShibInfo(Student student) {
        // Read new info
        StudentShibInfo newInfo = readShibInfo();
        
        // Get old info from database
        DataStorage ds = DataStorage.getInstance();
        StudentShibInfo oldInfo = ds.getStudentShibInfo(student.getId());

        if (oldInfo == null) {
            //Logging.error(this, "No shibInfo found for user " + student.getId() + ". Creating.");
            ds.saveStudentShibInfo(newInfo);
        } else if (!newInfo.equals(oldInfo)) {
            // Update those fields that are not manually overridden
            //Logging.info(this, "student.getStudentNumber() (" + student.getStudentNumber() + ")");
            //Logging.info(this, "oldInfo.getStudentNumber() (" + oldInfo.getStudentNumber() + ")");

            if (student.getName() == null || student.getName().equals(oldInfo.getCommonName())) {
                student.setName(newInfo.getCommonName());
                ds.setStudentField(student.getId(), "name", newInfo.getCommonName());
            }

            if (student.getEmail() == null || student.getEmail().equals(oldInfo.getEmail())) {
                student.setEmail(newInfo.getEmail());
                ds.setStudentField(student.getId(), "email", newInfo.getEmail());
            }

            if (student.getStudentNumber() == null || student.getStudentNumber().equals(oldInfo.getStudentNumber())) {
                student.setStudentNumber(newInfo.getStudentNumber());
                ds.setStudentField(student.getId(), "studentNumber", newInfo.getStudentNumber());
            }

            if (student.getLanguage() == null || student.getLanguage().equals(oldInfo.getLang())) {
                student.setLanguage(newInfo.getLang());
                ds.setStudentField(student.getId(), "lang", newInfo.getLang());
            }

            // Save shibInfo to database
            //Logging.info(this, "Updating shib info for " + student.getId());
            ds.updateStudentShibInfo(newInfo);
        }
    }

    /**
     * Determines the user's organization using the GET parameters and uses the
     * information to form the right WAYF url. The returned string begins with
     * "/Shibboleth.sso/".
     */
    private String determineShibUrl() {
        // Handle the GET parameters
        String oId = getRequestGlobals().getRequest().getParameter("oid");
        String cId = getRequestGlobals().getRequest().getParameter("cid");
        Organization organization = null;

        try {
            if (oId != null && oId.length() > 0) {
                organization = DataStorage.getInstance().getOrganization(Integer.parseInt(oId));
            }
        } catch (NumberFormatException e) {
            Logging.warn(this, "Invalid course id as a parameter for LoginShib page: " + cId);
        }

        // Form the URL
        String shibUrl = "/Shibboleth.sso/";
        String idp = null;

        if (organization != null) {
            idp = organization.getShibId();
        }

        if (idp != null && idp.length() > 0) {
            shibUrl += idp;
        } else {
            shibUrl += "wayf";
        }

        shibUrl += "?target=https://"
                + System.getProperty("https.server")
                + "/app%3Fservice=page%26page=LoginShib";

        // Do we have the cid? We don't want to lose it.
        if (cId != null && cId.length() > 0) {
            Logging.info(this, "appending cid");
            shibUrl += "%26cid=" + cId;
        }

        return shibUrl;
    }

    /**
     * Parses the studentnumber provided by the IdP.
     * The raw value is in form "urn:mace:terena.org:schac:personalUniqueCode:fi:tkk.fi:student:12345S"
     */
    private String parseShibStudentNumber(String rawString) {
        // For testing
        //rawString = "urn:mace:terena.org:schac:personalUniqueCode:fi:tkk.fi:student:12345S";
        if (rawString == null) {
            return null;
        }

        String[] array = rawString.split(":");

        if (array.length < 2) {
            return null;
        } else {
            return array[array.length - 1];
        }
    }

    public String getRefreshUrl() {
        if (isInformationComplete()) {
            return "<meta HTTP-EQUIV=\"refresh\" CONTENT=\"0;URL=http://" + System.getProperty("https.server") + "/app?service=page&page=Main\"/>";
        } else {
            return "<meta HTTP-EQUIV=\"refresh\" CONTENT=\"0;URL=http://" + System.getProperty("https.server") + "/app?service=page&page=Settings\"/>";
        }
    }
    
}

